from typing import Any, ClassVar

import pandas as pd
from sqlalchemy import UniqueConstraint
from sqlalchemy.orm import validates

from ixmp4 import db
from ixmp4.data import types
from ixmp4.data.abstract import optimization as abstract

from .. import Column, base


class Table(base.BaseModel):
    # NOTE: These might be mixin-able, but would require some abstraction
    NotFound: ClassVar = abstract.Table.NotFound
    NotUnique: ClassVar = abstract.Table.NotUnique
    DeletionPrevented: ClassVar = abstract.Table.DeletionPrevented

    # constrained_to_indexsets: ClassVar[list[str] | None] = None

    name: types.String = db.Column(db.String(255), nullable=False, unique=False)
    data: types.JsonDict = db.Column(db.JsonType, nullable=False, default={})

    columns: types.Mapped[list["Column"]] = db.relationship()

    # NOTE: Could create a RunMixin for all optimization models, see
    # https://docs.sqlalchemy.org/en/20/orm/declarative_mixins.html#mixing-in-relationships
    run__id: types.Integer = db.Column(
        db.Integer, db.ForeignKey("run.id"), nullable=False, index=True
    )

    __table_args__ = (UniqueConstraint(name, "run__id"),)

    # NOTE: This could probably also be a Mixin across almost all models
    created_at: types.DateTime = db.Column(db.DateTime, nullable=True)
    created_by: types.String = db.Column(db.String(255), nullable=True)

    def collect_indexsets_to_check(self) -> dict[str, Any]:
        """Creates a {key:value} dict from linked Column.names and their
        IndexSet.elements."""
        collection: dict[str, Any] = {}
        for column in self.columns:
            collection[column.name] = column.indexset.elements
        return collection

    @validates("data")
    def validate_data(self, key, data: dict[str, Any]):
        data_frame: pd.DataFrame = pd.DataFrame.from_dict(data)
        # TODO for all of the following, we might want to create unique exceptions
        # Could me make both more specific by specifiying missing/extra columns?
        if len(data_frame.columns) < len(self.columns):
            raise ValueError(
                f"Data is missing for some Columns! \n Data: {data} \n "
                f"Columns: {[column.name for column in self.columns]}"
            )
        elif len(data_frame.columns) > len(self.columns):
            raise ValueError(
                f"Trying to add data to unknown Columns! \n Data: {data} \n "
                f"Columns: {[column.name for column in self.columns]}"
            )

        # We could make this more specific maybe by pointing to the missing values
        if data_frame.isna().any(axis=None):
            raise ValueError(
                "Table.data is missing values, please make sure it does "
                "not contain None or NaN, either!"
            )
        # We can make this more specific e.g. highlighting all duplicate rows via
        # pd.DataFrame.duplicated(keep="False")
        if data_frame.value_counts().max() > 1:
            raise ValueError("Table.data contains duplicate rows!")

        # Can we make this more specific? Iterating over columns; if any is False,
        # return its name or something?
        limited_to_indexsets = self.collect_indexsets_to_check()
        if not data_frame.isin(limited_to_indexsets).all(axis=None):
            raise ValueError(
                "Table.data contains values that are not allowed as per the IndexSets "
                "and Columns it is constrained to!"
            )

        return data_frame.to_dict(orient="list")



import logging
import logging.config
import os
from pathlib import Path
from typing import Literal

from httpx import ConnectError
from pydantic import Field, HttpUrl, field_validator
from pydantic_settings import BaseSettings, SettingsConfigDict

from ixmp4.core.exceptions import InvalidCredentials

from .auth import AnonymousAuth, ManagerAuth
from .credentials import Credentials
from .manager import ManagerConfig
from .toml import TomlConfig
from .user import local_user

logger = logging.getLogger(__name__)

here = Path(__file__).parent


class Settings(BaseSettings):
    mode: Literal["production"] | Literal["development"] | Literal[
        "debug"
    ] = "production"
    storage_directory: Path = Field(
        "~/.local/share/ixmp4/", json_schema_extra={"env": "ixmp4_dir"}
    )
    secret_hs256: str = "default_secret_hs256"
    migration_db_uri: str = "sqlite:///./run/db.sqlite"
    manager_url: HttpUrl = Field("https://api.manager.ece.iiasa.ac.at/v1")
    managed: bool = True
    max_page_size: int = 10_000
    default_page_size: int = 5_000
    client_default_upload_chunk_size: int = 10_000
    client_max_concurrent_requests: int = Field(2, lt=4)
    client_max_request_retries: int = Field(3)
    client_backoff_factor: int = Field(5)
    client_timeout: int = Field(30)
    model_config = SettingsConfigDict(env_prefix="ixmp4_", extra="allow")

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.storage_directory.mkdir(parents=True, exist_ok=True)

        self.database_dir = self.storage_directory / "databases"
        self.database_dir.mkdir(exist_ok=True)

        self.log_dir = self.storage_directory / "log"
        self.log_dir.mkdir(exist_ok=True)

        self.configure_logging(self.mode)

        self._credentials = None
        self._toml = None
        self._default_auth = None
        self._manager = None

        logger.debug(f"Settings loaded: {self}")

    @property
    def credentials(self):
        if self._credentials is None:
            self.load_credentials()
        return self._credentials

    @property
    def default_credentials(self):
        try:
            return self.credentials.get("default")
        except KeyError:
            pass

    @property
    def toml(self):
        if self._toml is None:
            self.load_toml_config()
        return self._toml

    @property
    def default_auth(self):
        if self._default_auth is None:
            self.get_auth()
        return self._default_auth

    @property
    def manager(self):
        if self._manager is None:
            self.load_manager_config()
        return self._manager

    def load_credentials(self):
        credentials_config = self.storage_directory / "credentials.toml"
        credentials_config.touch()
        self._credentials = Credentials(credentials_config)

    def get_auth(self):
        if self.default_credentials is not None:
            try:
                self._default_auth = ManagerAuth(
                    *self.default_credentials, str(self.manager_url)
                )
                logger.info(
                    f"Connecting as user '{self._default_auth.get_user().username}'."
                )
            except InvalidCredentials:
                logger.warning(f"Invalid credentials for {self.manager_url}.")
            except ConnectError:
                logger.warning(f"Unable to connect to {self.manager_url}.")

        else:
            self._default_auth = AnonymousAuth()

    def load_manager_config(self):
        self._manager = ManagerConfig(
            str(self.manager_url), self.default_auth, remote=True
        )

    def load_toml_config(self):
        if self.default_auth is not None:
            toml_user = self.default_auth.get_user()
            if not toml_user.is_authenticated:
                toml_user = local_user
        else:  # if no connection to manager
            toml_user = local_user

        toml_config = self.storage_directory / "platforms.toml"
        toml_config.touch()
        self._toml = TomlConfig(toml_config, toml_user)

    @field_validator("storage_directory")
    def expand_user(cls, v):
        # translate ~/asdf into /home/user/asdf
        return Path.expanduser(v)

    def configure_logging(self, config: str):
        access_file = self.log_dir / "access.log"
        debug_file = self.log_dir / "debug.log"
        error_file = self.log_dir / "error.log"
        os.environ.setdefault("IXMP4_ACCESS_LOG", str(access_file.absolute()))
        os.environ.setdefault("IXMP4_DEBUG_LOG", str(debug_file.absolute()))
        os.environ.setdefault("IXMP4_ERROR_LOG", str(error_file.absolute()))

        logging_config = here / f"logging/{config}.conf"
        logging.config.fileConfig(logging_config, disable_existing_loggers=False)

    def check_credentials(self):
        if self.default_credentials is not None:
            username, password = self.default_credentials
            ManagerAuth(username, password, str(self.manager_url))



from datetime import datetime

import pandas as pd

from ixmp4.core.base import BaseFacade, BaseModelFacade
from ixmp4.data.abstract import Docs as DocsModel
from ixmp4.data.abstract import Unit as UnitModel


def to_dimensionless(name):
    if name == "dimensionless":
        raise ValueError(
            "Unit name 'dimensionless' is reserved, use an empty string '' instead."
        )
    return "dimensionless" if name == "" else name


class Unit(BaseModelFacade):
    _model: UnitModel
    NotUnique = UnitModel.NotUnique
    NotFound = UnitModel.NotFound
    DeletionPrevented = UnitModel.DeletionPrevented

    @property
    def id(self) -> int:
        return self._model.id

    @property
    def name(self) -> str:
        return self._model.name

    @property
    def created_at(self) -> datetime | None:
        return self._model.created_at

    @property
    def created_by(self) -> str | None:
        return self._model.created_by

    def delete(self):
        self.backend.units.delete(self._model.id)

    @property
    def docs(self):
        try:
            return self.backend.units.docs.get(self.id).description
        except DocsModel.NotFound:
            return None

    @docs.setter
    def docs(self, description):
        if description is None:
            self.backend.units.docs.delete(self.id)
        else:
            self.backend.units.docs.set(self.id, description)

    @docs.deleter
    def docs(self):
        try:
            self.backend.units.docs.delete(self.id)
        # TODO: silently failing
        except DocsModel.NotFound:
            return None

    def __str__(self) -> str:
        return f"<Unit {self.id} name={self.name}>"


class UnitRepository(BaseFacade):
    def create(
        self,
        name: str,
    ) -> Unit:
        name = to_dimensionless(name)
        if name.strip() == "":
            raise ValueError("Using a space-only unit name is not allowed.")

        model = self.backend.units.create(name)
        return Unit(_backend=self.backend, _model=model)

    def delete(self, x: Unit | int | str):
        if isinstance(x, Unit):
            id = x.id
        elif isinstance(x, int):
            id = x
        elif isinstance(x, str):
            model = self.backend.units.get(x)
            id = model.id
        else:
            raise TypeError("Invalid argument: Must be `Unit`, `int` or `str`.")

        self.backend.units.delete(id)

    def get(self, name: str) -> Unit:
        model = self.backend.units.get(to_dimensionless(name))
        return Unit(_backend=self.backend, _model=model)

    def list(self, name: str | None = None) -> list[Unit]:
        units = self.backend.units.list(name=name)
        return [Unit(_backend=self.backend, _model=u) for u in units]

    def tabulate(self, name: str | None = None) -> pd.DataFrame:
        return self.backend.units.tabulate(name=name)

    def _get_unit_id(self, unit: str) -> int | None:
        if unit is None:
            return None
        elif isinstance(unit, str):
            obj = self.backend.units.get(unit)
            return obj.id
        else:
            raise ValueError(f"Invalid reference to unit: {unit}")

    def get_docs(self, name: str) -> str | None:
        unit_id = self._get_unit_id(name)
        if unit_id is None:
            return None
        try:
            return self.backend.units.docs.get(dimension_id=unit_id).description
        except DocsModel.NotFound:
            return None

    def set_docs(self, name: str, description: str | None) -> str | None:
        if description is None:
            self.delete_docs(name=name)
            return None
        unit_id = self._get_unit_id(name)
        if unit_id is None:
            return None
        return self.backend.units.docs.set(
            dimension_id=unit_id, description=description
        ).description

    def delete_docs(self, name: str) -> None:
        # TODO: this function is failing silently, which we should avoid
        unit_id = self._get_unit_id(name)
        if unit_id is None:
            return None
        try:
            self.backend.units.docs.delete(dimension_id=unit_id)
            return None
        except DocsModel.NotFound:
            return None



from typing import Any, Iterable, Protocol

import pandas as pd

from ixmp4.data import types

from .. import base
from ..docs import DocsRepository
from .column import Column


class Table(base.BaseModel, Protocol):
    """Table data model."""

    name: types.String
    """Unique name of the Table."""
    data: types.JsonDict
    """Data stored in the Table."""
    columns: types.Mapped[list[Column]]
    """Data specifying this Table's Columns."""

    run__id: types.Integer
    "Foreign unique integer id of a run."

    created_at: types.DateTime
    "Creation date/time. TODO"
    created_by: types.String
    "Creator. TODO"

    def __str__(self) -> str:
        return f"<Table {self.id} name={self.name}>"


class TableRepository(
    base.Creator,
    base.Retriever,
    base.Enumerator,
    Protocol,
):
    docs: DocsRepository

    def create(
        self,
        run_id: int,
        name: str,
        constrained_to_indexsets: list[str],
        column_names: list[str] | None = None,
    ) -> Table:
        """Creates a Table.

        Each column of the Table needs to be constrained to an existing
        :class:ixmp4.data.abstract.optimization.IndexSet. These are specified by name
        and per default, these will be the column names. They can be overwritten by
        specifying `column_names`, which needs to specify a unique name for each column.

        Parameters
        ----------
        run_id : int
            The id of the :class:`ixmp4.data.abstract.Run` for which this Table is
            defined.
        name : str
            The unique name of the Table.
        constrained_to_indexsets : list[str]
            List of :class:`ixmp4.data.abstract.optimization.IndexSet` names that define
            the allowed contents of the Table's columns.
        column_names: list[str] | None = None
            Optional list of names to use as column names. If given, overwrites the
            names inferred from `constrained_to_indexsets`.

        Raises
        ------
        :class:`ixmp4.data.abstract.optimization.Table.NotUnique`:
            If the Table with `name` already exists for the Run with `run_id`.
        ValueError
            If `column_names` are not unique or not enough names are given.

        Returns
        -------
        :class:`ixmp4.data.abstract.optimization.Table`:
            The created Table.
        """
        ...

    def get(self, run_id: int, name: str) -> Table:
        """Retrieves a Table.

        Parameters
        ----------
        run_id : int
            The id of the :class:`ixmp4.data.abstract.Run` for which this Table is
            defined.
        name : str
            The name of the Table.

        Raises
        ------
        :class:`ixmp4.data.abstract.optimization.Table.NotFound`:
            If the Table with `name` does not exist.

        Returns
        -------
        :class:`ixmp4.data.abstract.optimization.Table`:
            The retrieved Table.
        """
        ...

    def get_by_id(self, id: int) -> Table:
        """Retrieves a Table by its id.

        Parameters
        ----------
        id : int
            Unique integer id.

        Raises
        ------
        :class:`ixmp4.data.abstract.optimization.Table.NotFound`.
            If the Table with `id` does not exist.

        Returns
        -------
        :class:`ixmp4.data.abstract.optimization.Table`:
            The retrieved Table.
        """
        ...

    def list(self, *, name: str | None = None, **kwargs) -> Iterable[Table]:
        r"""Lists Tables by specified criteria.

        Parameters
        ----------
        name : str
            The name of a Table. If supplied only one result will be returned.
        # TODO: Update kwargs
        \*\*kwargs: any
            More filter parameters as specified in
            `ixmp4.data.db.iamc.variable.filters.VariableFilter`.

        Returns
        -------
        Iterable[:class:`ixmp4.data.abstract.optimization.Table`]:
            List of Tables.
        """
        ...

    def tabulate(self, *, name: str | None = None, **kwargs) -> pd.DataFrame:
        r"""Tabulate Tables by specified criteria.

        Parameters
        ----------
        name : str
            The name of a Table. If supplied only one result will be returned.
        # TODO: Update kwargs
        \*\*kwargs: any
            More filter parameters as specified in
            `ixmp4.data.db.iamc.variable.filters.VariableFilter`.

        Returns
        -------
        :class:`pandas.DataFrame`:
            A data frame with the columns:
                - id
                - name
                - data
                - run__id
                - created_at
                - created_by
        """
        ...

    # TODO Once present, state how to check which IndexSets are linked and which values
    # they permit
    def add_data(self, table_id: int, data: dict[str, Any] | pd.DataFrame) -> None:
        r"""Adds data to a Table.

        The data will be validated with the linked constrained
        :class:`ixmp4.data.abstract.optimization.IndexSet`s. For that, `data.keys()`
        must correspond to the names of the Table's columns. Each column can only
        contain values that are in the linked `IndexSet.elements`. Each row of entries
        must be unique. No values can be missing, `None`, or `NaN`. If `data.keys()`
        contains names already present in `Table.data`, existing values will be
        overwritten.

        Parameters
        ----------
        table_id : int
            The id of the :class:`ixmp4.data.abstract.optimization.Table`.
        data : dict[str, Any] | pandas.DataFrame
            The data to be added.

        Raises
        ------
        ValueError:
            - If values are missing, `None`, or `NaN`
            - If values are not allowed based on constraints to `Indexset`s
            - If rows are not unique

        Returns
        -------
        None
        """
        ...



from ixmp4.data.db.docs import BaseDocsRepository, docs_model

from .model import Table


class TableDocsRepository(BaseDocsRepository):
    model_class = docs_model(Table)  # TableDocs
    dimension_model_class = Table



from typing import Mapping

from ixmp4 import db
from ixmp4.data import types
from ixmp4.data.db.iamc.measurand import Measurand
from ixmp4.data.db.region import Region
from ixmp4.data.db.timeseries import TimeSeries as BaseTimeSeries

from .. import base


class TimeSeries(BaseTimeSeries, base.BaseModel):
    __table_args__ = (db.UniqueConstraint("run__id", "region__id", "measurand__id"),)

    region__id: types.Integer = db.Column(
        db.Integer, db.ForeignKey("region.id"), nullable=False, index=True
    )
    region: types.Mapped[Region] = db.relationship(
        "Region", backref="metadata", foreign_keys=[region__id], lazy="select"
    )

    measurand__id: types.Integer = db.Column(
        db.Integer, db.ForeignKey("iamc_measurand.id"), nullable=False, index=True
    )
    measurand: types.Mapped[Measurand] = db.relationship(
        "Measurand", backref="metadata", foreign_keys=[measurand__id], lazy="select"
    )

    @property
    def parameters(self) -> Mapping:
        return {
            "region": self.region.name,
            "unit": self.measurand.unit.name,
            "variable": self.measurand.variable.name,
        }



from datetime import datetime
from typing import ClassVar, List

import pandas as pd
from pydantic import StrictFloat, StrictInt, StrictStr

from ixmp4.data import abstract

from .. import base
from ..docs import Docs, DocsRepository


class IndexSet(base.BaseModel):
    NotFound: ClassVar = abstract.IndexSet.NotFound
    NotUnique: ClassVar = abstract.IndexSet.NotUnique
    DeletionPrevented: ClassVar = abstract.IndexSet.DeletionPrevented

    id: int
    name: str
    elements: StrictFloat | StrictInt | list[
        StrictFloat | StrictInt | StrictStr
    ] | StrictStr | None
    run__id: int

    created_at: datetime | None
    created_by: str | None


class IndexSetDocsRepository(DocsRepository):
    model_class = Docs
    prefix = "docs/optimization/indexsets/"


class IndexSetRepository(
    base.Creator[IndexSet],
    base.Retriever[IndexSet],
    base.Enumerator[IndexSet],
    abstract.IndexSetRepository,
):
    model_class = IndexSet
    prefix = "optimization/indexsets/"

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.docs = IndexSetDocsRepository(self.backend)

    def create(
        self,
        run_id: int,
        name: str,
    ) -> IndexSet:
        return super().create(run_id=run_id, name=name)

    def get(self, run_id: int, name: str) -> IndexSet:
        return super().get(name=name, run_id=run_id)

    def enumerate(self, **kwargs) -> list[IndexSet] | pd.DataFrame:
        return super().enumerate(**kwargs)

    def list(self, **kwargs) -> list[IndexSet]:
        return super()._list(json=kwargs)

    def tabulate(self, **kwargs) -> pd.DataFrame:
        return super()._tabulate(json=kwargs)

    def add_elements(
        self,
        indexset_id: int,
        elements: StrictFloat
        | StrictInt
        | List[StrictFloat | StrictInt | StrictStr]
        | StrictStr,
    ) -> None:
        kwargs = {"indexset_id": indexset_id, "elements": elements}
        self._request("PATCH", self.prefix + str(indexset_id) + "/", json=kwargs)



from typing import ClassVar

from ixmp4.data.abstract import optimization as abstract

from .. import base
from .indexset import IndexSet


class Column(base.BaseModel):
    NotFound: ClassVar = abstract.Column.NotFound
    NotUnique: ClassVar = abstract.Column.NotUnique
    DeletionPrevented: ClassVar = abstract.Column.DeletionPrevented

    id: int
    name: str
    dtype: str
    table__id: int
    indexset: IndexSet
    constrained_to_indexset: int
    unique: bool



from fastapi import APIRouter, Body, Depends, Query

from ixmp4.data import api
from ixmp4.data.backend.db import SqlAlchemyBackend as Backend
from ixmp4.data.db.model.filter import ModelFilter

from . import deps
from .base import BaseModel, EnumerationOutput, Pagination
from .decorators import autodoc

router: APIRouter = APIRouter(
    prefix="/models",
    tags=["models"],
)


class ModelInput(BaseModel):
    name: str


@autodoc
@router.patch("/", response_model=EnumerationOutput[api.Model])
def query(
    filter: ModelFilter = Body(None),
    table: bool | None = Query(False),
    pagination: Pagination = Depends(),
    backend: Backend = Depends(deps.get_backend),
):
    return EnumerationOutput(
        results=backend.models.paginate(
            _filter=filter,
            limit=pagination.limit,
            offset=pagination.offset,
            table=bool(table),
        ),
        total=backend.models.count(_filter=filter),
        pagination=pagination,
    )


@autodoc
@router.post("/", response_model=api.Model)
def create(
    model: ModelInput,
    backend: Backend = Depends(deps.get_backend),
):
    return backend.models.create(**model.model_dump())



from typing import Protocol

from ixmp4.data import types

from . import base


class Docs(base.BaseModel, Protocol):
    """Abstract documentation model for one object of any dimension."""

    description: types.String
    "Description of the dimension object."
    dimension__id: types.Integer
    "Foreign unique integer id of the object in the dimension's table."
    dimension: types.Mapped
    "The documented object."

    # This doesn't work since each dimension has a different self.dimension object as
    # of now
    # def __str__(self) -> str:
    #    return (
    #        f"<Documentation for {self.dimension} {self.dimension.name}>"
    #    )


# TODO: adjust all type hints once things work
class DocsRepository(
    base.Retriever,
    base.Deleter,
    base.Enumerator,
    Protocol,
):
    def get(self, dimension_id: int) -> Docs:
        """Retrieve the documentation of an object of any dimension.

        Parameters
        ----------
        dimension_id : int
            The unique integer id of the object.

        Raises
        ------
        :class:`ixmp4.data.abstract.Docs.NotFound`:
            If the documentation for the object with `dimension_id` does not exist.

        Returns
        -------
        :class:`ixmp4.data.abstract.Docs`
            The object's documentation.
        """
        ...

    def set(self, dimension_id: int, description: str) -> Docs:
        """Sets the documentation for an object of any dimension.

        Parameters
        ----------
        dimension_id : int
            The id of the object in its dimension's table.
        description : str
            Description of the object.

        Returns
        -------
        :class:`ixmp4.data.abstract.Docs`
            The object's documentation.
        """
        ...

    def delete(self, dimension_id: int) -> None:
        """Deletes a dimension object's documentation.
        Warning: No recovery of deleted data shall be possible via ixmp
        after the execution of this function.

        Parameters
        ----------
        dimension_id : int
            The unique id of the object whose documentation should be deleted in its
            dimension's table.

        Raises
        ------
        :class:`ixmp4.data.abstract.Docs.NotFound`:
            If the documentation for the object with `dimension_id` does not exist.
        """
        ...

    def list(
        self,
        *,
        dimension_id: int | None = None,
    ) -> list[Docs]:
        """Lists documentations.

        Parameters
        ----------
        dimension_id : int
            The id of an object in any dimension. If supplied only one result will be
            returned.

        Returns
        -------
            Iterable[:class:`ixmp4.data.abstract.Docs`] : List of documentations.
        """
        ...



# type: ignore
"""Initial Migration

Revision ID: c71efc396d2b
Revises:
Create Date: 2023-04-26 15:37:46.677955

"""
import sqlalchemy as sa
from alembic import op

# Revision identifiers, used by Alembic.
revision = "c71efc396d2b"
down_revision = None
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "iamc_variable",
        sa.Column("name", sa.String(length=255), nullable=False),
        sa.Column("created_at", sa.DateTime(), nullable=True),
        sa.Column("created_by", sa.String(length=255), nullable=True),
        sa.Column(
            "id",
            sa.Integer(),
            sa.Identity(always=False, on_null=True, start=1, increment=1),
            nullable=False,
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_iamc_variable")),
        sa.UniqueConstraint("name", name=op.f("uq_iamc_variable_name")),
    )
    op.create_table(
        "model",
        sa.Column("name", sa.String(length=255), nullable=False),
        sa.Column("created_at", sa.DateTime(), nullable=True),
        sa.Column("created_by", sa.String(length=255), nullable=True),
        sa.Column(
            "id",
            sa.Integer(),
            sa.Identity(always=False, on_null=True, start=1, increment=1),
            nullable=False,
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_model")),
        sa.UniqueConstraint("name", name=op.f("uq_model_name")),
    )
    op.create_table(
        "region",
        sa.Column("name", sa.String(length=1023), nullable=False),
        sa.Column("hierarchy", sa.String(length=1023), nullable=False),
        sa.Column("created_at", sa.DateTime(), nullable=True),
        sa.Column("created_by", sa.String(length=255), nullable=True),
        sa.Column(
            "id",
            sa.Integer(),
            sa.Identity(always=False, on_null=True, start=1, increment=1),
            nullable=False,
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_region")),
        sa.UniqueConstraint("name", name=op.f("uq_region_name")),
    )
    op.create_table(
        "scenario",
        sa.Column("name", sa.String(length=255), nullable=False),
        sa.Column("created_at", sa.DateTime(), nullable=True),
        sa.Column("created_by", sa.String(length=255), nullable=True),
        sa.Column(
            "id",
            sa.Integer(),
            sa.Identity(always=False, on_null=True, start=1, increment=1),
            nullable=False,
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_scenario")),
        sa.UniqueConstraint("name", name=op.f("uq_scenario_name")),
    )
    op.create_table(
        "unit",
        sa.Column("name", sa.String(length=255), nullable=False),
        sa.Column("created_at", sa.DateTime(), nullable=True),
        sa.Column("created_by", sa.String(length=255), nullable=True),
        sa.Column(
            "id",
            sa.Integer(),
            sa.Identity(always=False, on_null=True, start=1, increment=1),
            nullable=False,
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_unit")),
        sa.UniqueConstraint("name", name=op.f("uq_unit_name")),
    )
    op.create_table(
        "iamc_measurand",
        sa.Column("variable__id", sa.Integer(), nullable=False),
        sa.Column("unit__id", sa.Integer(), nullable=False),
        sa.Column("created_at", sa.DateTime(), nullable=True),
        sa.Column("created_by", sa.String(length=255), nullable=True),
        sa.Column(
            "id",
            sa.Integer(),
            sa.Identity(always=False, on_null=True, start=1, increment=1),
            nullable=False,
        ),
        sa.ForeignKeyConstraint(
            ["unit__id"],
            ["unit.id"],
            name=op.f("fk_iamc_measurand_unit__id_unit"),
        ),
        sa.ForeignKeyConstraint(
            ["variable__id"],
            ["iamc_variable.id"],
            name=op.f("fk_iamc_measurand_variable__id_iamc_variable"),
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_iamc_measurand")),
        sa.UniqueConstraint(
            "variable__id",
            "unit__id",
            name=op.f("uq_iamc_measurand_variable__id_unit__id"),
        ),
    )
    with op.batch_alter_table("iamc_measurand", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_iamc_measurand_unit__id"),
            ["unit__id"],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f("ix_iamc_measurand_variable__id"),
            ["variable__id"],
            unique=False,
        )

    op.create_table(
        "iamc_variable_docs",
        sa.Column("description", sa.Text(), nullable=False),
        sa.Column("dimension__id", sa.Integer(), nullable=True),
        sa.Column(
            "id",
            sa.Integer(),
            sa.Identity(always=False, on_null=True, start=1, increment=1),
            nullable=False,
        ),
        sa.ForeignKeyConstraint(
            ["dimension__id"],
            ["iamc_variable.id"],
            name=op.f("fk_iamc_variable_docs_dimension__id_iamc_variable"),
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_iamc_variable_docs")),
        sa.UniqueConstraint(
            "dimension__id", name=op.f("uq_iamc_variable_docs_dimension__id")
        ),
    )
    op.create_table(
        "model_docs",
        sa.Column("description", sa.Text(), nullable=False),
        sa.Column("dimension__id", sa.Integer(), nullable=True),
        sa.Column(
            "id",
            sa.Integer(),
            sa.Identity(always=False, on_null=True, start=1, increment=1),
            nullable=False,
        ),
        sa.ForeignKeyConstraint(
            ["dimension__id"],
            ["model.id"],
            name=op.f("fk_model_docs_dimension__id_model"),
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_model_docs")),
        sa.UniqueConstraint("dimension__id", name=op.f("uq_model_docs_dimension__id")),
    )
    op.create_table(
        "region_docs",
        sa.Column("description", sa.Text(), nullable=False),
        sa.Column("dimension__id", sa.Integer(), nullable=True),
        sa.Column(
            "id",
            sa.Integer(),
            sa.Identity(always=False, on_null=True, start=1, increment=1),
            nullable=False,
        ),
        sa.ForeignKeyConstraint(
            ["dimension__id"],
            ["region.id"],
            name=op.f("fk_region_docs_dimension__id_region"),
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_region_docs")),
        sa.UniqueConstraint("dimension__id", name=op.f("uq_region_docs_dimension__id")),
    )
    op.create_table(
        "run",
        sa.Column("model__id", sa.Integer(), nullable=False),
        sa.Column("scenario__id", sa.Integer(), nullable=False),
        sa.Column("version", sa.Integer(), nullable=False),
        sa.Column("is_default", sa.Boolean(), nullable=False),
        sa.Column(
            "id",
            sa.Integer(),
            sa.Identity(always=False, on_null=True, start=1, increment=1),
            nullable=False,
        ),
        sa.ForeignKeyConstraint(
            ["model__id"], ["model.id"], name=op.f("fk_run_model__id_model")
        ),
        sa.ForeignKeyConstraint(
            ["scenario__id"],
            ["scenario.id"],
            name=op.f("fk_run_scenario__id_scenario"),
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_run")),
        sa.UniqueConstraint(
            "model__id",
            "scenario__id",
            "version",
            name=op.f("uq_run_model__id_scenario__id_version"),
        ),
    )
    with op.batch_alter_table("run", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_run_model__id"), ["model__id"], unique=False
        )
        batch_op.create_index(
            batch_op.f("ix_run_scenario__id"), ["scenario__id"], unique=False
        )

    op.create_table(
        "scenario_docs",
        sa.Column("description", sa.Text(), nullable=False),
        sa.Column("dimension__id", sa.Integer(), nullable=True),
        sa.Column(
            "id",
            sa.Integer(),
            sa.Identity(always=False, on_null=True, start=1, increment=1),
            nullable=False,
        ),
        sa.ForeignKeyConstraint(
            ["dimension__id"],
            ["scenario.id"],
            name=op.f("fk_scenario_docs_dimension__id_scenario"),
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_scenario_docs")),
        sa.UniqueConstraint(
            "dimension__id", name=op.f("uq_scenario_docs_dimension__id")
        ),
    )
    op.create_table(
        "unit_docs",
        sa.Column("description", sa.Text(), nullable=False),
        sa.Column("dimension__id", sa.Integer(), nullable=True),
        sa.Column(
            "id",
            sa.Integer(),
            sa.Identity(always=False, on_null=True, start=1, increment=1),
            nullable=False,
        ),
        sa.ForeignKeyConstraint(
            ["dimension__id"],
            ["unit.id"],
            name=op.f("fk_unit_docs_dimension__id_unit"),
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_unit_docs")),
        sa.UniqueConstraint("dimension__id", name=op.f("uq_unit_docs_dimension__id")),
    )
    op.create_table(
        "iamc_timeseries",
        sa.Column("region__id", sa.Integer(), nullable=False),
        sa.Column("measurand__id", sa.Integer(), nullable=False),
        sa.Column("run__id", sa.Integer(), nullable=False),
        sa.Column(
            "id",
            sa.Integer(),
            sa.Identity(always=False, on_null=True, start=1, increment=1),
            nullable=False,
        ),
        sa.ForeignKeyConstraint(
            ["measurand__id"],
            ["iamc_measurand.id"],
            name=op.f("fk_iamc_timeseries_measurand__id_iamc_measurand"),
        ),
        sa.ForeignKeyConstraint(
            ["region__id"],
            ["region.id"],
            name=op.f("fk_iamc_timeseries_region__id_region"),
        ),
        sa.ForeignKeyConstraint(
            ["run__id"],
            ["run.id"],
            name=op.f("fk_iamc_timeseries_run__id_run"),
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_iamc_timeseries")),
        sa.UniqueConstraint(
            "run__id",
            "region__id",
            "measurand__id",
            name=op.f("uq_iamc_timeseries_run__id_region__id_measurand__id"),
        ),
    )
    with op.batch_alter_table("iamc_timeseries", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_iamc_timeseries_measurand__id"),
            ["measurand__id"],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f("ix_iamc_timeseries_region__id"),
            ["region__id"],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f("ix_iamc_timeseries_run__id"), ["run__id"], unique=False
        )

    op.create_table(
        "runmetaentry",
        sa.Column("run__id", sa.Integer(), nullable=False),
        sa.Column("key", sa.String(length=1023), nullable=False),
        sa.Column("type", sa.String(length=20), nullable=False),
        sa.Column("value_int", sa.Integer(), nullable=True),
        sa.Column("value_str", sa.String(length=1023), nullable=True),
        sa.Column("value_float", sa.Float(), nullable=True),
        sa.Column("value_bool", sa.Boolean(), nullable=True),
        sa.Column(
            "id",
            sa.Integer(),
            sa.Identity(always=False, on_null=True, start=1, increment=1),
            nullable=False,
        ),
        sa.ForeignKeyConstraint(
            ["run__id"], ["run.id"], name=op.f("fk_runmetaentry_run__id_run")
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_runmetaentry")),
        sa.UniqueConstraint("run__id", "key", name=op.f("uq_runmetaentry_run__id_key")),
    )
    with op.batch_alter_table("runmetaentry", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_runmetaentry_run__id"), ["run__id"], unique=False
        )

    op.create_table(
        "iamc_datapoint_oracle",
        sa.Column("value", sa.Float(), nullable=True),
        sa.Column("type", sa.String(length=255), nullable=False),
        sa.Column("step_category", sa.String(length=1023), nullable=True),
        sa.Column("step_year", sa.Integer(), nullable=True),
        sa.Column("step_datetime", sa.DateTime(), nullable=True),
        sa.Column("time_series__id", sa.Integer(), nullable=False),
        sa.Column(
            "id",
            sa.Integer(),
            sa.Identity(always=False, on_null=True, start=1, increment=1),
            nullable=False,
        ),
        sa.ForeignKeyConstraint(
            ["time_series__id"],
            ["iamc_timeseries.id"],
            name=op.f("fk_iamc_datapoint_oracle_time_series__id_iamc_timeseries"),
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_iamc_datapoint_oracle")),
        sa.UniqueConstraint(
            "time_series__id",
            "step_year",
            "step_category",
            "step_datetime",
            name=op.f(
                "uq_iamc_datapoint_oracle_time_series__id_step_year_step_category_step_datetime"
            ),
        ),
    )
    with op.batch_alter_table("iamc_datapoint_oracle", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_iamc_datapoint_oracle_step_category"),
            ["step_category"],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f("ix_iamc_datapoint_oracle_step_datetime"),
            ["step_datetime"],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f("ix_iamc_datapoint_oracle_step_year"),
            ["step_year"],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f("ix_iamc_datapoint_oracle_time_series__id"),
            ["time_series__id"],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f("ix_iamc_datapoint_oracle_type"), ["type"], unique=False
        )

    op.create_table(
        "iamc_datapoint_universal",
        sa.Column("value", sa.Float(), nullable=True),
        sa.Column("type", sa.String(length=255), nullable=False),
        sa.Column("step_category", sa.String(length=1023), nullable=True),
        sa.Column("step_year", sa.Integer(), nullable=True),
        sa.Column("step_datetime", sa.DateTime(), nullable=True),
        sa.Column("time_series__id", sa.Integer(), nullable=False),
        sa.Column(
            "id",
            sa.Integer(),
            sa.Identity(always=False, on_null=True, start=1, increment=1),
            nullable=False,
        ),
        sa.ForeignKeyConstraint(
            ["time_series__id"],
            ["iamc_timeseries.id"],
            name=op.f("fk_iamc_datapoint_universal_time_series__id_iamc_timeseries"),
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_iamc_datapoint_universal")),
        sa.UniqueConstraint(
            "time_series__id",
            "step_datetime",
            name=op.f("uq_iamc_datapoint_universal_time_series__id_step_datetime"),
        ),
        sa.UniqueConstraint(
            "time_series__id",
            "step_year",
            "step_category",
            name=op.f(
                "uq_iamc_datapoint_universal_time_series__id_step_year_step_category"
            ),
        ),
    )
    with op.batch_alter_table("iamc_datapoint_universal", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_iamc_datapoint_universal_step_category"),
            ["step_category"],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f("ix_iamc_datapoint_universal_step_datetime"),
            ["step_datetime"],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f("ix_iamc_datapoint_universal_step_year"),
            ["step_year"],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f("ix_iamc_datapoint_universal_time_series__id"),
            ["time_series__id"],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f("ix_iamc_datapoint_universal_type"),
            ["type"],
            unique=False,
        )

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("iamc_datapoint_universal", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_iamc_datapoint_universal_type"))
        batch_op.drop_index(batch_op.f("ix_iamc_datapoint_universal_time_series__id"))
        batch_op.drop_index(batch_op.f("ix_iamc_datapoint_universal_step_year"))
        batch_op.drop_index(batch_op.f("ix_iamc_datapoint_universal_step_datetime"))
        batch_op.drop_index(batch_op.f("ix_iamc_datapoint_universal_step_category"))

    op.drop_table("iamc_datapoint_universal")
    with op.batch_alter_table("iamc_datapoint_oracle", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_iamc_datapoint_oracle_type"))
        batch_op.drop_index(batch_op.f("ix_iamc_datapoint_oracle_time_series__id"))
        batch_op.drop_index(batch_op.f("ix_iamc_datapoint_oracle_step_year"))
        batch_op.drop_index(batch_op.f("ix_iamc_datapoint_oracle_step_datetime"))
        batch_op.drop_index(batch_op.f("ix_iamc_datapoint_oracle_step_category"))

    op.drop_table("iamc_datapoint_oracle")
    with op.batch_alter_table("runmetaentry", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_runmetaentry_run__id"))

    op.drop_table("runmetaentry")
    with op.batch_alter_table("iamc_timeseries", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_iamc_timeseries_run__id"))
        batch_op.drop_index(batch_op.f("ix_iamc_timeseries_region__id"))
        batch_op.drop_index(batch_op.f("ix_iamc_timeseries_measurand__id"))

    op.drop_table("iamc_timeseries")
    op.drop_table("unit_docs")
    op.drop_table("scenario_docs")
    with op.batch_alter_table("run", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_run_scenario__id"))
        batch_op.drop_index(batch_op.f("ix_run_model__id"))

    op.drop_table("run")
    op.drop_table("region_docs")
    op.drop_table("model_docs")
    op.drop_table("iamc_variable_docs")
    with op.batch_alter_table("iamc_measurand", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_iamc_measurand_variable__id"))
        batch_op.drop_index(batch_op.f("ix_iamc_measurand_unit__id"))

    op.drop_table("iamc_measurand")
    op.drop_table("unit")
    op.drop_table("scenario")
    op.drop_table("region")
    op.drop_table("model")
    op.drop_table("iamc_variable")
    # ### end Alembic commands ###



from ixmp4.data.db import filters as base
from ixmp4.data.db.optimization.table import Table
from ixmp4.db import filters, utils

from .model import Column


# NB: This is currently not in use, but would be the basis to filter Columns by
# corresponding Tables
class OptimizationTableFilter(
    base.OptimizationTableFilter, metaclass=filters.FilterMeta
):
    def join(self, exc, session=None):
        if not utils.is_joined(exc, Table):
            exc = exc.join(Table, onclause=Column.table__id == Table.id)
        return exc


class OptimizationColumnFilter(
    base.OptimizationColumnFilter, metaclass=filters.FilterMeta
):
    def join(self, exc, session=None):
        return exc



from typing import ClassVar

from ixmp4.db import filters

from .. import Run, Scalar, Unit


class OptimizationScalarFilter(filters.BaseFilter, metaclass=filters.FilterMeta):
    id: filters.Id
    name: filters.String
    run__id: filters.Integer = filters.Field(None, alias="run_id")
    unit__id: filters.Integer = filters.Field(None, alias="unit_id")

    sqla_model: ClassVar[type] = Scalar

    def join(self, exc, **kwargs):
        exc = exc.join(Run, onclause=Scalar.run__id == Run.id)
        exc = exc.join(Unit, onclause=Scalar.unit__id == Unit.id)
        return exc



from typing import ClassVar

from ixmp4.db import filters

from .. import IndexSet, Run


class OptimizationIndexSetFilter(filters.BaseFilter, metaclass=filters.FilterMeta):
    id: filters.Id
    name: filters.String
    run__id: filters.Integer = filters.Field(None, alias="run_id")

    sqla_model: ClassVar[type] = IndexSet

    def join(self, exc, **kwargs):
        exc = exc.join(Run, onclause=IndexSet.run__id == Run.id)
        return exc



from typing import Protocol

from ixmp4.data import types

from .. import base
from .indexset import IndexSet


# TODO: standardize docstrings (run/Run)
class Column(base.BaseModel, Protocol):
    """Column data model."""

    name: types.String
    """Unique name of the Column."""
    dtype: types.String
    """Type of the Column's data."""
    table__id: types.Integer
    """Foreign unique integer id of a Table."""
    indexset: types.Mapped[IndexSet]
    """Associated IndexSet."""
    constrained_to_indexset: types.Integer
    """Foreign unique integer id of an IndexSet."""

    unique: types.Boolean
    """Boolean to determine whether data in this Column must contribute to Uniqueness
    of Keys."""

    def __str__(self) -> str:
        return f"<Column {self.id} name={self.name}>"



"""
This module contains sqlalchemy database models and repositories.
"""

from .base import BaseModel, BaseRepository
from .docs import BaseDocsRepository, docs_model
from .iamc import (  # AnnualDataPoint,; SubAnnualDataPoint,; CategoricalDataPoint,
    DataPoint,
    DataPointRepository,
    Measurand,
    MeasurandRepository,
    TimeSeries,
    TimeSeriesRepository,
    UniversalDataPoint,
    Variable,
    VariableRepository,
)
from .meta import RunMetaEntry, RunMetaEntryRepository
from .model import Model, ModelRepository
from .optimization import (
    Column,
    ColumnRepository,
    IndexSet,
    IndexSetRepository,
    Scalar,
    ScalarRepository,
    Table,
    TableRepository,
)
from .region import Region, RegionRepository
from .run import Run, RunRepository
from .scenario import Scenario, ScenarioRepository
from .unit import Unit, UnitRepository



from datetime import datetime
from typing import ClassVar

import pandas as pd

from ixmp4.core.base import BaseFacade, BaseModelFacade
from ixmp4.data.abstract import Docs as DocsModel
from ixmp4.data.abstract import Variable as VariableModel


class Variable(BaseModelFacade):
    _model: VariableModel
    NotFound: ClassVar = VariableModel.NotFound
    NotUnique: ClassVar = VariableModel.NotUnique

    @property
    def id(self) -> int:
        return self._model.id

    @property
    def name(self) -> str:
        return self._model.name

    @property
    def created_at(self) -> datetime | None:
        return self._model.created_at

    @property
    def created_by(self) -> str | None:
        return self._model.created_by

    @property
    def docs(self):
        try:
            return self.backend.iamc.variables.docs.get(self.id).description
        except DocsModel.NotFound:
            return None

    @docs.setter
    def docs(self, description):
        if description is None:
            self.backend.iamc.variables.docs.delete(self.id)
        else:
            self.backend.iamc.variables.docs.set(self.id, description)

    @docs.deleter
    def docs(self):
        try:
            self.backend.iamc.variables.docs.delete(self.id)
        # TODO: silently failing
        except DocsModel.NotFound:
            return None

    def __str__(self) -> str:
        return f"<Variable {self.id} name={self.name}>"


class VariableRepository(BaseFacade):
    def create(
        self,
        name: str,
    ) -> Variable:
        model = self.backend.iamc.variables.create(name)
        return Variable(_backend=self.backend, _model=model)

    def get(self, name: str) -> Variable:
        model = self.backend.iamc.variables.get(name)
        return Variable(_backend=self.backend, _model=model)

    def list(self, name: str | None = None) -> list[Variable]:
        variables = self.backend.iamc.variables.list(name=name)
        return [Variable(_backend=self.backend, _model=v) for v in variables]

    def tabulate(self, name: str | None = None) -> pd.DataFrame:
        return self.backend.iamc.variables.tabulate(name=name)

    def _get_variable_id(self, variable: str) -> int | None:
        if variable is None:
            return None
        elif isinstance(variable, str):
            obj = self.backend.iamc.variables.get(variable)
            return obj.id
        else:
            raise ValueError(f"Invalid reference to variable: {variable}")

    def get_docs(self, name: str) -> str | None:
        variable_id = self._get_variable_id(name)
        if variable_id is None:
            return None
        try:
            return self.backend.iamc.variables.docs.get(
                dimension_id=variable_id
            ).description
        except DocsModel.NotFound:
            return None

    def set_docs(self, name: str, description: str | None) -> str | None:
        if description is None:
            self.delete_docs(name=name)
            return None
        variable_id = self._get_variable_id(name)
        if variable_id is None:
            return None
        return self.backend.iamc.variables.docs.set(
            dimension_id=variable_id, description=description
        ).description

    def delete_docs(self, name: str) -> None:
        # TODO: this function is failing silently, which we should avoid
        variable_id = self._get_variable_id(name)
        if variable_id is None:
            return None
        try:
            self.backend.iamc.variables.docs.delete(dimension_id=variable_id)
            return None
        except DocsModel.NotFound:
            return None



# flake8: noqa
from ..base import BaseModel as RootBaseModel
from ..base import (
    BulkDeleter,
    BulkUpserter,
    Creator,
    Deleter,
    Enumerator,
    Lister,
    Retriever,
    Selecter,
    Tabulator,
)


class BaseModel(RootBaseModel):
    __abstract__ = True
    table_prefix = "optimization_"



from ixmp4.data.db.docs import BaseDocsRepository, docs_model

from .model import Column


class ColumnDocsRepository(BaseDocsRepository):
    model_class = docs_model(Column)  # ColumnDocs
    dimension_model_class = Column



from functools import wraps
from typing import TYPE_CHECKING, Callable, Protocol

from ixmp4.core.exceptions import Forbidden, ProgrammingError

if TYPE_CHECKING:
    from ..backend.db import SqlAlchemyBackend


class Guardable(Protocol):
    # NOTE: Eager checking for api backends may be desirable
    # at some point
    backend: "SqlAlchemyBackend"


def guard(access: str) -> Callable:
    if access not in ["edit", "manage", "view"]:
        raise ProgrammingError("Guard access must be 'edit', 'manage' or 'view'.")

    def decorator(func):
        @wraps(func)
        def guarded_func(self: Guardable, *args, **kwargs):
            if self.backend.auth_context is not None:
                if access == "view" and self.backend.auth_context.is_viewable:
                    return func(self, *args, **kwargs)
                elif access == "edit" and self.backend.auth_context.is_editable:
                    return func(self, *args, **kwargs)
                elif access == "manage" and self.backend.auth_context.is_managed:
                    return func(self, *args, **kwargs)

                raise Forbidden(
                    f"Function '{func.__name__}': '{access}' access denied due to "
                    "insufficient permissions."
                )
            else:
                # ignoring authorization since no auth context is set
                return func(self, *args, **kwargs)

        return guarded_func

    return decorator



from typing import Any, Iterable

import pandas as pd

from ixmp4 import db
from ixmp4.data.abstract import optimization as abstract
from ixmp4.data.auth.decorators import guard

from .. import ColumnRepository, base
from .docs import TableDocsRepository
from .model import Table


class TableRepository(
    base.Creator[Table],
    base.Retriever[Table],
    base.Enumerator[Table],
    abstract.TableRepository,
):
    model_class = Table

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.docs = TableDocsRepository(*args, **kwargs)
        self.columns = ColumnRepository(*args, **kwargs)

        from .filter import OptimizationTableFilter

        self.filter_class = OptimizationTableFilter

    def _add_column(
        self,
        run_id: int,
        table_id: int,
        column_name: str,
        indexset_name: str,
        **kwargs,
    ) -> None:
        r"""Adds a Column to a Table.

        Parameters
        ----------
        run_id : int
            The id of the :class:`ixmp4.data.abstract.Run` for which the
            :class:`ixmp4.data.abstract.optimization.Table` is defined.
        table_id : int
            The id of the :class:`ixmp4.data.abstract.optimization.Table`.
        column_name : str
            The name of the Column, which must be unique in connection with the names of
            :class:`ixmp4.data.abstract.Run` and
            :class:`ixmp4.data.abstract.optimization.Table`.
        indexset_name : str
            The name of the :class:`ixmp4.data.abstract.optimization.IndexSet` the
            Column will be linked to.
        \*\*kwargs: any
            Keyword arguments to be passed to
            :func:`ixmp4.data.abstract.optimization.Column.create`.
        """
        indexset = self.backend.optimization.indexsets.get(
            run_id=run_id, name=indexset_name
        )
        self.columns.create(
            table_id=table_id,
            name=column_name,
            dtype=pd.Series(indexset.elements).dtype.name,
            constrained_to_indexset=indexset.id,
            unique=True,
            **kwargs,
        )

    def add(
        self,
        run_id: int,
        name: str,
    ) -> Table:
        table = Table(name=name, run__id=run_id, **self.get_creation_info())
        self.session.add(table)

        return table

    @guard("view")
    def get(self, run_id: int, name: str) -> Table:
        exc = db.select(Table).where((Table.name == name) & (Table.run__id == run_id))
        try:
            return self.session.execute(exc).scalar_one()
        except db.NoResultFound:
            raise Table.NotFound

    @guard("view")
    def get_by_id(self, id: int) -> Table:
        obj = self.session.get(self.model_class, id)

        if obj is None:
            raise Table.NotFound(id=id)

        return obj

    @guard("edit")
    def create(
        self,
        run_id: int,
        name: str,
        constrained_to_indexsets: list[str],
        column_names: list[str] | None = None,
        **kwargs,
    ) -> Table:
        # Convert to list to avoid enumerate() splitting strings to letters
        if isinstance(constrained_to_indexsets, str):
            constrained_to_indexsets = list(constrained_to_indexsets)
        if column_names and len(column_names) != len(constrained_to_indexsets):
            raise ValueError(
                "`constrained_to_indexsets` and `column_names` not equal in length! "
                "Please provide the same number of entries for both!"
            )
        # TODO: activate something like this if each column must be indexed by a unique
        # indexset
        # if len(constrained_to_indexsets) != len(set(constrained_to_indexsets)):
        #     raise ValueError("Each dimension must be constrained to a unique indexset!") # noqa
        if column_names and len(column_names) != len(set(column_names)):
            raise ValueError("The given `column_names` are not unique!")

        table = super().create(
            run_id=run_id,
            name=name,
            **kwargs,
        )
        for i, name in enumerate(constrained_to_indexsets):
            self._add_column(
                run_id=run_id,
                table_id=table.id,
                column_name=column_names[i] if column_names else name,
                indexset_name=name,
            )

        return table

    @guard("view")
    def list(self, *args, **kwargs) -> Iterable[Table]:
        return super().list(*args, **kwargs)

    @guard("view")
    def tabulate(self, *args, **kwargs) -> pd.DataFrame:
        return super().tabulate(*args, **kwargs)

    @guard("edit")
    def add_data(self, table_id: int, data: dict[str, Any] | pd.DataFrame) -> None:
        if isinstance(data, dict):
            data = pd.DataFrame.from_dict(data=data)
        table = self.get_by_id(id=table_id)

        table.data = pd.concat([pd.DataFrame.from_dict(table.data), data]).to_dict(
            orient="list"
        )

        self.session.add(table)
        self.session.commit()



# type: ignore
"""Create optimization_table table

Revision ID: 081bbda6bb7b
Revises: 97ba231770e2
Create Date: 2024-04-12 10:00:13.662611

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

# Revision identifiers, used by Alembic.
revision = "081bbda6bb7b"
down_revision = "97ba231770e2"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "optimization_table",
        sa.Column("name", sa.String(length=255), nullable=False),
        sa.Column(
            "data",
            sa.JSON().with_variant(
                postgresql.JSONB(astext_type=sa.Text()), "postgresql"
            ),
            nullable=False,
        ),
        sa.Column("run__id", sa.Integer(), nullable=False),
        sa.Column("created_at", sa.DateTime(), nullable=True),
        sa.Column("created_by", sa.String(length=255), nullable=True),
        sa.Column(
            "id",
            sa.Integer(),
            sa.Identity(always=False, on_null=True, start=1, increment=1),
            nullable=False,
        ),
        sa.ForeignKeyConstraint(
            ["run__id"], ["run.id"], name=op.f("fk_optimization_table_run__id_run")
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_optimization_table")),
        sa.UniqueConstraint(
            "name", "run__id", name=op.f("uq_optimization_table_name_run__id")
        ),
    )
    with op.batch_alter_table("optimization_table", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_optimization_table_run__id"), ["run__id"], unique=False
        )

    op.create_table(
        "optimization_column",
        sa.Column("name", sa.String(length=255), nullable=False),
        sa.Column("dtype", sa.String(length=255), nullable=False),
        sa.Column("table__id", sa.Integer(), nullable=False),
        sa.Column("constrained_to_indexset", sa.Integer(), nullable=False),
        sa.Column("unique", sa.Boolean(), nullable=False),
        sa.Column(
            "id",
            sa.Integer(),
            sa.Identity(always=False, on_null=True, start=1, increment=1),
            nullable=False,
        ),
        sa.ForeignKeyConstraint(
            ["constrained_to_indexset"],
            ["optimization_indexset.id"],
            name=op.f(
                "fk_optimization_column_constrained_to_indexset_optimization_indexset"
            ),
        ),
        sa.ForeignKeyConstraint(
            ["table__id"],
            ["optimization_table.id"],
            name=op.f("fk_optimization_column_table__id_optimization_table"),
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_optimization_column")),
        sa.UniqueConstraint(
            "name", "table__id", name=op.f("uq_optimization_column_name_table__id")
        ),
    )
    with op.batch_alter_table("optimization_column", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_optimization_column_constrained_to_indexset"),
            ["constrained_to_indexset"],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f("ix_optimization_column_table__id"), ["table__id"], unique=False
        )

    op.create_table(
        "optimization_table_docs",
        sa.Column("description", sa.Text(), nullable=False),
        sa.Column("dimension__id", sa.Integer(), nullable=True),
        sa.Column(
            "id",
            sa.Integer(),
            sa.Identity(always=False, on_null=True, start=1, increment=1),
            nullable=False,
        ),
        sa.ForeignKeyConstraint(
            ["dimension__id"],
            ["optimization_table.id"],
            name=op.f("fk_optimization_table_docs_dimension__id_optimization_table"),
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_optimization_table_docs")),
        sa.UniqueConstraint(
            "dimension__id", name=op.f("uq_optimization_table_docs_dimension__id")
        ),
    )
    op.create_table(
        "optimization_column_docs",
        sa.Column("description", sa.Text(), nullable=False),
        sa.Column("dimension__id", sa.Integer(), nullable=True),
        sa.Column(
            "id",
            sa.Integer(),
            sa.Identity(always=False, on_null=True, start=1, increment=1),
            nullable=False,
        ),
        sa.ForeignKeyConstraint(
            ["dimension__id"],
            ["optimization_column.id"],
            name=op.f("fk_optimization_column_docs_dimension__id_optimization_column"),
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_optimization_column_docs")),
        sa.UniqueConstraint(
            "dimension__id", name=op.f("uq_optimization_column_docs_dimension__id")
        ),
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table("optimization_column_docs")
    op.drop_table("optimization_table_docs")
    with op.batch_alter_table("optimization_column", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_optimization_column_table__id"))
        batch_op.drop_index(
            batch_op.f("ix_optimization_column_constrained_to_indexset")
        )

    op.drop_table("optimization_column")
    with op.batch_alter_table("optimization_table", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_optimization_table_run__id"))

    op.drop_table("optimization_table")
    # ### end Alembic commands ###



from typing import Generic, Mapping, Protocol, TypeVar

import pandas as pd

from ixmp4.data import types

from .. import base


class TimeSeries(base.BaseModel, Protocol):
    """Time series data model."""

    run__id: types.Integer
    "Unique run id."
    parameters: Mapping
    "A set of parameter values for the time series."

    def __str__(self) -> str:
        return f"<TimeSeries {self.id} parameters={self.parameters}>"


ModelType = TypeVar("ModelType", bound=TimeSeries)


class TimeSeriesRepository(
    base.Creator,
    base.Retriever,
    base.Enumerator,
    base.BulkUpserter,
    Protocol,
    Generic[ModelType],
):
    def create(self, run_id: int, parameters: Mapping) -> ModelType:
        """Retrieves a time series.

        Parameters
        ----------
        run_id : int
            Unique run id.
        parameters : Mapping
            A set of parameter values for the time series.

        Raises
        ------
        :class:`ixmp4.data.abstract.TimeSeries.NotUnique`.

        Returns
        -------
        :class:`ixmp4.data.base.TimeSeries`:
            The retrieved time series.
        """
        ...

    def get(self, run_id: int, parameters: Mapping) -> ModelType:
        """Retrieves a time series.

        Parameters
        ----------
        run_id : int
            Unique run id.
        parameters : Mapping
            A set of parameter values for the time series.

        Raises
        ------
        :class:`ixmp4.data.abstract.TimeSeries.NotFound`.
            If the time series with `run_id` and `parameters` does not exist.

        Returns
        -------
        :class:`ixmp4.data.base.TimeSeries`:
            The retrieved time series.
        """
        ...

    def get_by_id(self, id: int) -> ModelType:
        """Retrieves a time series by it's id.

        Parameters
        ----------
        id : int
            Unique integer id.

        Raises
        ------
        :class:`ixmp4.data.abstract.TimeSeries.NotFound`.
            If the time series with `id` does not exist.

        Returns
        -------
        :class:`ixmp4.data.base.TimeSeries`:
            The retrieved time series.
        """
        ...

    def get_or_create(self, run_id: int, parameters: Mapping) -> ModelType:
        """Tries to retrieve a time series and creates it if it was not found.

        Parameters
        ----------
        run_id : int
            Unique run id.
        parameters : Mapping
            A set of parameter values for the time series.

        Returns
        -------
        :class:`ixmp4.data.base.TimeSeries`:
            The retrieved or created time series.
        """
        try:
            return self.get(run_id, parameters)
        except TimeSeries.NotFound:
            return self.create(run_id, parameters)

    def list(
        self,
        **kwargs,
    ) -> list[ModelType]:
        r"""Lists time series by specified criteria.

        Parameters
        ----------
            \*\*kwargs: any
            Filter parameters as specified in
            :class:`ixmp4.data.db.iamc.timeseries.filter.TimeSeriesFilter`.

        Returns
        -------
        Iterable[:class:`ixmp4.data.base.TimeSeries`]:
            List of time series.
        """
        ...

    def tabulate(
        self, *, join_parameters: bool | None = False, **kwargs
    ) -> pd.DataFrame:
        r"""Tabulate time series by specified criteria.

        Parameters
        ----------
        join_parameters : bool | None
            If set to `True` the resulting data frame will include
            parameter columns as values instead of foreign key id's.
         \*\*kwargs: any
            Filter parameters as specified in
            :class:`ixmp4.data.db.iamc.timeseries.filter.TimeSeriesFilter`.

        Returns
        -------
        :class:`pandas.DataFrame`:
            A data frame with the columns:
                - id
                - run__id
                - ... parameter id columns
                  Or:
                - ... parameter value columns
        """
        ...

    def bulk_upsert(
        self, df: pd.DataFrame, create_related: bool | None = False
    ) -> None:
        """Looks which time series in the supplied data frame already exist, and
        inserts new ones.

        Parameters
        ----------
        df : :class:`pandas.DataFrame`
            A data frame with the columns:
                - id
                - run__id

                if `create_related` = False:
                    - ... parameter id columns

                else:
                    - ... parameter value columns (i.e `region` column with value
                      "World" instead of `region__id` column with value "1".)

        create_related : bool
            Creates related database entries with value data supplied in `df`.
        """
        ...



from typing import Any

from fastapi import APIRouter, Body, Depends, Query

from ixmp4.data import api
from ixmp4.data.backend.db import SqlAlchemyBackend as Backend
from ixmp4.data.db.optimization.table.filter import OptimizationTableFilter

from .. import deps
from ..base import BaseModel, EnumerationOutput, Pagination
from ..decorators import autodoc

router: APIRouter = APIRouter(
    prefix="/tables",
    tags=["optimization", "tables"],
)


class TableCreateInput(BaseModel):
    run_id: int
    name: str
    constrained_to_indexsets: list[str]
    column_names: list[str] | None


class DataInput(BaseModel):
    data: dict[str, Any]


@autodoc
@router.get("/{id}/", response_model=api.Table)
def get_by_id(
    id: int,
    backend: Backend = Depends(deps.get_backend),
):
    return backend.optimization.tables.get_by_id(id)


@autodoc
@router.patch("/", response_model=EnumerationOutput[api.Table])
def query(
    filter: OptimizationTableFilter = Body(OptimizationTableFilter(id=None, name=None)),
    table: bool = Query(False),
    pagination: Pagination = Depends(),
    backend: Backend = Depends(deps.get_backend),
):
    return EnumerationOutput(
        results=backend.optimization.tables.paginate(
            _filter=filter,
            limit=pagination.limit,
            offset=pagination.offset,
            table=bool(table),
        ),
        total=backend.optimization.tables.count(_filter=filter),
        pagination=pagination,
    )


@autodoc
@router.patch("/{table_id}/data/")
def add_data(
    table_id: int,
    data: DataInput,
    backend: Backend = Depends(deps.get_backend),
):
    return backend.optimization.tables.add_data(table_id=table_id, **data.model_dump())


@autodoc
@router.post("/", response_model=api.Table)
def create(
    table: TableCreateInput,
    backend: Backend = Depends(deps.get_backend),
):
    return backend.optimization.tables.create(**table.model_dump())



import enum
from typing import Protocol

import pandas as pd

from ixmp4.data import types

from .. import base


class DataPoint(base.BaseModel, Protocol):
    """Data point data model."""

    time_series__id: types.Integer
    "Foreign unique integer id of the associated time series."
    value: types.Float
    "Value of the data point."
    type: types.String
    "Type of data point either `ANNUAL`, `CATEGORICAL` or `DATETIME`."

    step_year: types.Integer | None
    "An integer time step required by data points of type `ANNUAL` or `CATEGORICAL`."
    step_category: types.String | None
    "A string category required by data points of type `CATEGORICAL`."
    step_datetime: types.DateTime | None
    "A datetime object required by data points of type `DATETIME`."

    class Type(str, enum.Enum):
        BASE = "BASE"
        ANNUAL = "ANNUAL"
        CATEGORICAL = "CATEGORICAL"
        DATETIME = "DATETIME"

    def __str__(self) -> str:
        return f"<Datapoint {self.id} type={self.type}>"


class DataPointRepository(
    base.Enumerator,
    base.BulkUpserter,
    base.BulkDeleter,
    Protocol,
):
    def list(
        self,
        *,
        join_parameters: bool | None = False,
        **kwargs,
    ) -> list[DataPoint]:
        """Lists data points by specified criteria.
        This method incurrs mentionable overhead compared to :meth:`tabulate`.

        Parameters
        ----------
        join_parameters : bool | None
            If set to `True` the resulting data frame will include parameter columns
            from the associated :class:`ixmp4.data.base.TimeSeries`.
        kwargs : Any
            Additional key word arguments. Any left over kwargs will be used as filters.

        Returns
        -------
        Iterable[:class:`ixmp4.data.base.DataPoint`]:
            List of data points.
        """
        ...

    def tabulate(
        self,
        *,
        join_parameters: bool | None = False,
        join_runs: bool = False,
        **kwargs,
    ) -> pd.DataFrame:
        """Tabulates data points by specified criteria.

        Parameters
        ----------
        join_parameters : bool | None
            If set to `True` the resulting data frame will include parameter columns
            from the associated :class:`ixmp4.data.abstract.TimeSeries`.
        join_runs : bool
            If set to `True` the resulting data frame will include model & scenario name
            and version id of the associated Run.
        kwargs : Any
            Additional key word arguments. Any left over kwargs will be used as filters.

        Returns
        -------
        :class:`pandas.DataFrame`:
            A data frame with the columns:
                - id
                - time_series__id
                - value
                - type
                - step_year
                    if it contains data points of type `ANNUAL` or `CATEGORICAL`
                - step_category
                    if it contains data points of type `CATEGORICAL`
                - step_datetime
                    if it contains data points of type `DATETIME`
                - ... misc parameter columns if `join_parameters` is set to `True`
        """
        ...

    def bulk_upsert(self, df: pd.DataFrame, **kwargs) -> None:
        """Looks which data points in the supplied data frame already exists,
        updates those that have changed and inserts new ones.

        Parameters
        ----------
        df : :class:`pandas.DataFrame`
            A data frame with the columns:
                - time_series__id
                - value
                - type
                - step_year
                    if it contains data points of type `ANNUAL` or `CATEGORICAL`
                - step_category
                    if it contains data points of type `CATEGORICAL`
                - step_datetime
                    if it contains data points of type `DATETIME`
        """
        ...

    def bulk_delete(self, df: pd.DataFrame, **kwargs) -> None:
        """Deletes data points which match criteria in the supplied data frame.

        Parameters
        ----------
        df : :class:`pandas.DataFrame`
            A data frame with the columns:
                - time_series__id
                - type
                - step_year
                    if it contains data points of type `ANNUAL` or `CATEGORICAL`
                - step_category
                    if it contains data points of type `CATEGORICAL`
                - step_datetime
                    if it contains data points of type `DATETIME`
        """
        ...



from typing import ClassVar

from ixmp4.db import filters

from .. import RunMetaEntry


class RunMetaEntryFilter(filters.BaseFilter, metaclass=filters.FilterMeta):
    sqla_model: ClassVar[type] = RunMetaEntry

    id: filters.Id
    type: filters.String
    run__id: filters.Integer = filters.Field(None, alias="run_id")

    key: filters.String

    value_int: filters.Integer
    value_str: filters.String
    value_float: filters.Float
    value_bool: filters.Boolean

    def join(self, exc, **kwargs):
        return exc



from fastapi import APIRouter, Body, Depends, Query

from ixmp4.data import api
from ixmp4.data.backend.db import SqlAlchemyBackend as Backend
from ixmp4.data.db.optimization.scalar.filter import OptimizationScalarFilter

from .. import deps
from ..base import BaseModel, EnumerationOutput, Pagination
from ..decorators import autodoc

router: APIRouter = APIRouter(
    prefix="/scalars",
    tags=["optimization", "scalars"],
)


class ScalarCreateInput(BaseModel):
    name: str
    value: float
    unit_name: str
    run_id: int


class ScalarUpdateInput(BaseModel):
    value: float | None
    unit_id: int | None


@autodoc
@router.get("/{id}/", response_model=api.Scalar)
def get_by_id(
    id: int,
    backend: Backend = Depends(deps.get_backend),
):
    return backend.optimization.scalars.get_by_id(id)


@autodoc
@router.patch("/", response_model=EnumerationOutput[api.Scalar])
def query(
    filter: OptimizationScalarFilter = Body(
        OptimizationScalarFilter(id=None, name=None, unit__id=None)
    ),
    table: bool = Query(False),
    pagination: Pagination = Depends(),
    backend: Backend = Depends(deps.get_backend),
):
    return EnumerationOutput(
        results=backend.optimization.scalars.paginate(
            _filter=filter,
            limit=pagination.limit,
            offset=pagination.offset,
            table=bool(table),
        ),
        total=backend.optimization.scalars.count(_filter=filter),
        pagination=pagination,
    )


@autodoc
@router.patch("/{id}/", response_model=api.Scalar)
def update(
    id: int,
    scalar: ScalarUpdateInput,
    backend: Backend = Depends(deps.get_backend),
):
    return backend.optimization.scalars.update(id, **scalar.model_dump())


@autodoc
@router.post("/", response_model=api.Scalar)
def create(
    scalar: ScalarCreateInput,
    backend: Backend = Depends(deps.get_backend),
):
    return backend.optimization.scalars.create(**scalar.model_dump())



from ..docs import BaseDocsRepository, docs_model
from .model import Scenario


class ScenarioDocsRepository(BaseDocsRepository):
    model_class = docs_model(Scenario)  # ScenarioDocs
    dimension_model_class = Scenario



from ixmp4.data.auth.decorators import guard

from .. import base
from .docs import ColumnDocsRepository
from .model import Column


class ColumnRepository(
    base.Creator[Column],
    base.Retriever[Column],
):
    model_class = Column

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.docs = ColumnDocsRepository(*args, **kwargs)

        from .filter import OptimizationColumnFilter

        self.filter_class = OptimizationColumnFilter

    def add(
        self,
        table_id: int,
        name: str,
        dtype: str,
        constrained_to_indexset: str,
        unique: bool,
    ) -> Column:
        column = Column(
            table__id=table_id,
            name=name,
            dtype=dtype,
            constrained_to_indexset=constrained_to_indexset,
            unique=unique,
        )
        self.session.add(column)
        return column

    @guard("edit")
    def create(
        self,
        table_id: int,
        name: str,
        dtype: str,
        constrained_to_indexset: int,
        unique: bool,
        **kwargs,
    ) -> Column:
        """Creates a Column.

        Parameters
        ----------
        table_id : int
            The unique integer id of the :class:`ixmp4.data.abstract.optimization.Table`
            this Column belongs to.
        name : str
            The unique name of the Column.
        dtype : str
            The pandas-inferred type of the Column's data.
        constrained_to_indexset : int
            The id of an :class:`ixmp4.data.abstract.optimization.IndexSet`, which must
            contain all values used as entries in this Column.
        unique : bool
            A bool to determine whether entries in this Column should be considered for
            evaluating uniqueness of keys. Defaults to True.

        Raises
        ------
        :class:`ixmp4.data.abstract.optimization.Column.NotUnique`:
            If the Column with `name` already exists for the related
            :class:`ixmp4.data.abstract.optimization.Table`.

        Returns
        -------
        :class:`ixmp4.data.abstract.optimization.Column`:
            The created Column.
        """
        return super().create(
            table_id=table_id,
            name=name,
            dtype=dtype,
            constrained_to_indexset=constrained_to_indexset,
            unique=unique,
            **kwargs,
        )



from typing import ClassVar

from ixmp4 import db
from ixmp4.data import abstract, types
from ixmp4.data.db.model.model import Model
from ixmp4.data.db.optimization.indexset import IndexSet
from ixmp4.data.db.optimization.scalar import Scalar
from ixmp4.data.db.optimization.table import Table
from ixmp4.data.db.scenario.model import Scenario

from .. import base


class Run(base.BaseModel):
    NotFound: ClassVar = abstract.Run.NotFound
    NotUnique: ClassVar = abstract.Run.NotUnique
    DeletionPrevented: ClassVar = abstract.Run.DeletionPrevented

    NoDefaultVersion: ClassVar = abstract.Run.NoDefaultVersion

    __table_args__ = (db.UniqueConstraint("model__id", "scenario__id", "version"),)

    model__id: types.Integer = db.Column(
        db.ForeignKey("model.id"), nullable=False, index=True
    )
    model: types.Mapped[Model] = db.relationship(
        "Model",
        backref="run",
        foreign_keys=[model__id],
    )

    scenario__id: types.Integer = db.Column(
        db.ForeignKey("scenario.id"),
        nullable=False,
        index=True,
    )
    scenario: types.Mapped[Scenario] = db.relationship(
        "Scenario",
        backref="run",
        foreign_keys=[scenario__id],
    )

    indexsets: types.Mapped[list["IndexSet"]] = db.relationship()
    scalars: types.Mapped[list["Scalar"]] = db.relationship()
    tables: types.Mapped[list["Table"]] = db.relationship()

    version: types.Integer = db.Column(db.Integer, nullable=False)
    is_default: types.Boolean = db.Column(db.Boolean, default=False, nullable=False)



from ixmp4.data.db import filters as base
from ixmp4.data.db.run import Run
from ixmp4.data.db.unit import Unit
from ixmp4.db import filters, utils

from .model import Scalar


class OptimizationUnitFilter(base.UnitFilter, metaclass=filters.FilterMeta):
    def join(self, exc, **kwawrgs):
        if not utils.is_joined(exc, Unit):
            exc = exc.join(Unit, onclause=Scalar.unit__id == Unit.id)
        return exc


class RunFilter(base.RunFilter, metaclass=filters.FilterMeta):
    def join(self, exc, **kwargs):
        if not utils.is_joined(exc, Run):
            exc = exc.join(Run, onclause=Scalar.run__id == Run.id)
        return exc


class OptimizationScalarFilter(
    base.OptimizationScalarFilter, metaclass=filters.FilterMeta
):
    def join(self, exc, session=None):
        return exc



from typing import ClassVar

from ixmp4.db import filters, utils

from .. import Run, TimeSeries


class TimeSeriesFilter(filters.BaseFilter, metaclass=filters.FilterMeta):
    id: filters.Id

    sqla_model: ClassVar[type] = TimeSeries

    def join(self, exc, **kwargs):
        if not utils.is_joined(exc, TimeSeries):
            exc = exc.join(TimeSeries, onclause=TimeSeries.run__id == Run.id)
        return exc



import json
from pathlib import Path
from typing import Any

import toml

from ixmp4.core.exceptions import PlatformNotFound, PlatformNotUnique

from .base import Config, PlatformInfo
from .user import User


class TomlPlatformInfo(PlatformInfo):
    pass


class TomlConfig(Config):
    platforms: dict[str, TomlPlatformInfo]

    def __init__(self, toml_file: Path, user: User) -> None:
        self.path = toml_file
        self.user = user
        self.load()

    def load(self) -> None:
        dict_ = toml.load(self.path)
        list_: list[dict[str, Any]] = [{"name": k, **v} for k, v in dict_.items()]
        self.platforms = {x["name"]: TomlPlatformInfo(**x) for x in list_}

    def dump(self):
        obj = {}
        for c in self.platforms.values():
            dict_ = json.loads(c.model_dump_json())
            dict_.pop("user", None)
            name = dict_.pop("name")
            obj[name] = dict_

        f = self.path.open("w+")
        toml.dump(obj, f)

    def list_platforms(self) -> list[TomlPlatformInfo]:
        return list(self.platforms.values())

    def get_platform(self, key: str) -> TomlPlatformInfo:
        try:
            return self.platforms[key]
        except KeyError as e:
            raise PlatformNotFound(f"Platform '{key}' was not found.") from e

    def add_platform(self, name: str, dsn: str):
        try:
            self.get_platform(name)
        except PlatformNotFound:
            self.platforms[name] = TomlPlatformInfo(name=name, dsn=dsn)
            self.dump()
            return
        raise PlatformNotUnique(f"Platform '{name}' already exists, remove it first.")

    def remove_platform(self, key: str):
        try:
            del self.platforms[key]
        except KeyError as e:
            raise PlatformNotFound(f"Platform '{key}' was not found.") from e
        self.dump()



import pandas as pd

from ixmp4.data.abstract import DataPoint as DataPointModel


def substitute_type(df: pd.DataFrame, type: DataPointModel.Type | None = None):
    if "type" not in df.columns:
        # `type` given explicitly
        if type is not None:
            if type not in DataPointModel.Type:
                raise ValueError(f"Invalid data point type: {type}")
            df["type"] = type

        # pyam `data` format for annual timeseries data
        elif "year" in df.columns:
            df.rename(columns={"year": "step_year"}, inplace=True)
            df["type"] = DataPointModel.Type.ANNUAL

        else:
            raise ValueError("Required keyword argument `type` is `None`.")



from typing import Protocol

import pandas as pd

from ixmp4.data import types

from .. import base
from ..docs import DocsRepository


class Variable(base.BaseModel, Protocol):
    """Variable data model."""

    name: types.String
    "Unique name of the variable."

    created_at: types.DateTime
    "Creation date/time. TODO"
    created_by: types.String
    "Creator. TODO"

    def __str__(self) -> str:
        return f"<Variable {self.id} name={self.name}>"


class VariableRepository(
    base.Creator,
    base.Retriever,
    base.Enumerator,
    Protocol,
):
    docs: DocsRepository

    def create(self, name: str) -> Variable:
        """Creates a variable.

        Parameters
        ----------
        name : str
            The name of the variable.

        Raises
        ------
        :class:`ixmp4.data.abstract.Variable.NotUnique`:
            If the variable with `name` is not unique.

        Returns
        -------
        :class:`ixmp4.data.abstract.Variable`:
            The created variable.
        """
        ...

    def get(self, name: str) -> Variable:
        """Retrieves a variable.

        Parameters
        ----------
        name : str
            The unique name of the variable.

        Raises
        ------
        :class:`ixmp4.data.abstract.Variable.NotFound`:
            If the variable with `name` does not exist.

        Returns
        -------
        :class:`ixmp4.data.abstract.Variable`:
            The retrieved variable.
        """
        ...

    def list(self, *, name: str | None = None, **kwargs) -> list[Variable]:
        r"""Lists variables by specified criteria.

        Parameters
        ----------
        name : str
            The name of a variable. If supplied only one result will be returned.
        \*\*kwargs: any
            More filter parameters as specified in
            `ixmp4.data.db.iamc.variable.filters.VariableFilter`.

        Returns
        -------
        Iterable[:class:`ixmp4.data.abstract.Variable`]:
            List of variables.
        """
        ...

    def tabulate(self, *, name: str | None = None, **kwargs) -> pd.DataFrame:
        r"""Tabulate variables by specified criteria.

        Parameters
        ----------
        name : str
            The name of a variable. If supplied only one result will be returned.
        \*\*kwargs: any
            More filter parameters as specified in
            `ixmp4.data.db.iamc.variable.filters.VariableFilter`.

        Returns
        -------
        :class:`pandas.DataFrame`:
            A data frame with the columns:
                - id
                - name
        """
        ...



from ixmp4.data.db.docs import BaseDocsRepository, docs_model

from .model import IndexSet


class IndexSetDocsRepository(BaseDocsRepository):
    model_class = docs_model(IndexSet)  # IndexSetDocs
    dimension_model_class = IndexSet



from typing import ClassVar

import pandas as pd

from ixmp4.data import abstract

from . import base


class RunMetaEntry(base.BaseModel):
    NotFound: ClassVar = abstract.RunMetaEntry.NotFound
    NotUnique: ClassVar = abstract.RunMetaEntry.NotUnique
    DeletionPrevented: ClassVar = abstract.RunMetaEntry.DeletionPrevented

    id: int
    run__id: int
    key: str
    type: str
    value: abstract.StrictMetaValue


class RunMetaEntryRepository(
    base.Creator[RunMetaEntry],
    base.Retriever[RunMetaEntry],
    base.Deleter[RunMetaEntry],
    base.Enumerator[RunMetaEntry],
    base.BulkUpserter[RunMetaEntry],
    base.BulkDeleter[RunMetaEntry],
    abstract.RunMetaEntryRepository,
):
    model_class = RunMetaEntry
    prefix = "meta/"

    def create(
        self,
        run__id: int,
        key: str,
        value: abstract.MetaValue,
    ) -> RunMetaEntry:
        return super().create(run__id=run__id, key=key, value=value)

    def get(self, run__id: int, key: str) -> RunMetaEntry:
        return super().get(run_id=run__id, key=key)

    def delete(self, id: int) -> None:
        super().delete(id)

    def enumerate(self, **kwargs) -> list[RunMetaEntry] | pd.DataFrame:
        return super().enumerate(**kwargs)

    def list(self, join_run_index: bool | None = None, **kwargs) -> list[RunMetaEntry]:
        return super()._list(json=kwargs, params={"join_run_index": join_run_index})

    def tabulate(self, join_run_index: bool | None = None, **kwargs) -> pd.DataFrame:
        return super()._tabulate(json=kwargs, params={"join_run_index": join_run_index})

    def bulk_upsert(self, df: pd.DataFrame) -> None:
        super().bulk_upsert(df)

    def bulk_delete(self, df: pd.DataFrame) -> None:
        super().bulk_delete(df)



from contextlib import suppress
from pathlib import Path

import toml


class Credentials(object):
    credentials: dict

    def __init__(self, toml_file: Path) -> None:
        self.path = toml_file
        self.load()

    def load(self):
        self.credentials = toml.load(self.path)

    def dump(self):
        f = self.path.open("w+")
        toml.dump(self.credentials, f)

    def get(self, key: str) -> tuple[str, str]:
        c = self.credentials[key]
        return (c["username"], c["password"])

    def set(self, key: str, username: str, password: str):
        self.credentials[key] = {
            "username": username,
            "password": password,
        }
        self.dump()

    def clear(self, key: str):
        with suppress(KeyError):
            del self.credentials[key]
        self.dump()



from ixmp4.conf.base import PlatformInfo
from ixmp4.data.abstract import (
    DataPointRepository,
    IndexSetRepository,
    ModelRepository,
    RegionRepository,
    RunMetaEntryRepository,
    RunRepository,
    ScalarRepository,
    ScenarioRepository,
    TableRepository,
    TimeSeriesRepository,
    UnitRepository,
    VariableRepository,
)


class IamcSubobject(object):
    datapoints: DataPointRepository
    timeseries: TimeSeriesRepository
    variables: VariableRepository


class OptimizationSubobject(object):
    indexsets: IndexSetRepository
    scalars: ScalarRepository
    tables: TableRepository


class Backend(object):
    iamc: IamcSubobject
    info: PlatformInfo
    meta: RunMetaEntryRepository
    models: ModelRepository
    optimization: OptimizationSubobject
    regions: RegionRepository
    runs: RunRepository
    scenarios: ScenarioRepository
    units: UnitRepository

    def __init__(self, info: PlatformInfo) -> None:
        self.info = info
        self.iamc = IamcSubobject()
        self.optimization = OptimizationSubobject()

    def close(self) -> None:
        """Closes the connection to the database."""
        ...

    def __str__(self):
        return f"<Backend {self.info}>"



"""
IXMP4 provides a CLI for platform creation, deletion, etc.:

.. code:: bash

   # list all existing databases
   ixmp4 platforms list

   # run all migrations on all existing databases
   ixmp4 platforms upgrade

   # create a new database or register it
   ixmp4 platforms add <database-name> [--dsn sqlite://my/database/file.db]

   # delete a database
   ixmp4 platforms delete <database-name>

In development mode additional commands are available:

.. code:: bash

   # set the revision hash of all databases
   # without running migrations
   ixmp4 platforms stamp <revision-hash>

"""

from ixmp4.conf import settings
from ixmp4.conf.auth import BaseAuth
from ixmp4.core.exceptions import PlatformNotFound
from ixmp4.data.backend import Backend, RestBackend, SqlAlchemyBackend

from .iamc import PlatformIamcData
from .meta import MetaRepository
from .model import ModelRepository
from .region import RegionRepository
from .run import RunRepository
from .scenario import ScenarioRepository
from .unit import UnitRepository


class Platform(object):
    """A modeling platform instance as a connection to a data backend.
    Enables the manipulation of data via the `Run` class and `Repository` instances."""

    runs: RunRepository
    iamc: PlatformIamcData
    models: ModelRepository
    regions: RegionRepository
    scenarios: ScenarioRepository
    units: UnitRepository
    meta: MetaRepository

    backend: Backend
    """Provides a unified data interface for the platform.
    Using it directly is not recommended."""

    def __init__(
        self,
        name: str | None = None,
        _backend: Backend | None = None,
        _auth: BaseAuth | None = None,
    ) -> None:
        if name is not None:
            if name in settings.toml.platforms:
                config = settings.toml.get_platform(name)
            else:
                settings.check_credentials()
                if settings.manager is not None:
                    config = settings.manager.get_platform(name)
                else:
                    raise PlatformNotFound(f"Platform '{name}' was not found.")

            if config.dsn.startswith("http"):
                self.backend = RestBackend(config, auth=_auth)
            else:
                self.backend = SqlAlchemyBackend(config)  # type: ignore
        elif _backend is not None:
            self.backend = _backend
        else:
            raise TypeError("__init__() is missing required argument 'name'")

        self.runs = RunRepository(_backend=self.backend)
        self.iamc = PlatformIamcData(_backend=self.backend)
        self.models = ModelRepository(_backend=self.backend)
        self.regions = RegionRepository(_backend=self.backend)
        self.scenarios = ScenarioRepository(_backend=self.backend)
        self.units = UnitRepository(_backend=self.backend)
        self.meta = MetaRepository(_backend=self.backend)



from fastapi import APIRouter, Body, Depends, Query

from ixmp4.data import api
from ixmp4.data.backend.db import SqlAlchemyBackend as Backend
from ixmp4.data.db.scenario.filter import ScenarioFilter

from . import deps
from .base import BaseModel, EnumerationOutput, Pagination
from .decorators import autodoc

router: APIRouter = APIRouter(
    prefix="/scenarios",
    tags=["scenarios"],
)


class ScenarioInput(BaseModel):
    name: str


@autodoc
@router.patch("/", response_model=EnumerationOutput[api.Scenario])
def query(
    filter: ScenarioFilter = Body(None),
    table: bool | None = Query(False),
    pagination: Pagination = Depends(),
    backend: Backend = Depends(deps.get_backend),
):
    return EnumerationOutput(
        results=backend.scenarios.paginate(
            _filter=filter,
            limit=pagination.limit,
            offset=pagination.offset,
            table=bool(table),
        ),
        total=backend.scenarios.count(_filter=filter),
        pagination=pagination,
    )


@autodoc
@router.post("/", response_model=api.Scenario)
def create(
    scenario: ScenarioInput,
    backend: Backend = Depends(deps.get_backend),
):
    return backend.scenarios.create(**scenario.model_dump())



from datetime import datetime
from typing import Optional, Union

import pandas as pd

from ixmp4.data.abstract import Docs as DocsModel
from ixmp4.data.abstract import Region as RegionModel

from .base import BaseFacade, BaseModelFacade


class Region(BaseModelFacade):
    _model: RegionModel
    NotUnique = RegionModel.NotUnique
    NotFound = RegionModel.NotFound
    DeletionPrevented = RegionModel.DeletionPrevented

    @property
    def id(self) -> int:
        """Unique id."""
        return self._model.id

    @property
    def name(self) -> str:
        """Region name."""
        return self._model.name

    @property
    def hierarchy(self) -> str:
        """Region hierarchy."""
        return self._model.hierarchy

    @property
    def created_at(self) -> datetime | None:
        return self._model.created_at

    @property
    def created_by(self) -> str | None:
        return self._model.created_by

    def delete(self):
        """Deletes the region from the database."""
        self.backend.regions.delete(self._model.id)

    @property
    def docs(self):
        try:
            return self.backend.regions.docs.get(self.id).description
        except DocsModel.NotFound:
            return None

    @docs.setter
    def docs(self, description):
        if description is None:
            self.backend.regions.docs.delete(self.id)
        else:
            self.backend.regions.docs.set(self.id, description)

    @docs.deleter
    def docs(self):
        try:
            self.backend.regions.docs.delete(self.id)
        # TODO: silently failing
        except DocsModel.NotFound:
            return None

    def __str__(self) -> str:
        return f"<Region {self.id} name={self.name}>"


class RegionRepository(BaseFacade):
    def _get_region_id(self, region: Optional[Union[str, int, "Region"]]) -> int | None:
        if region is None:
            return None
        elif isinstance(region, str):
            obj = self.backend.regions.get(region)
            return obj.id
        elif isinstance(region, int):
            return region
        elif isinstance(region, Region):
            return region.id
        else:
            raise ValueError(f"Invalid reference to region: {region}")

    def create(
        self,
        name: str,
        hierarchy: str,
    ) -> Region:
        model = self.backend.regions.create(name, hierarchy)
        return Region(_backend=self.backend, _model=model)

    def get(self, name: str) -> Region:
        model = self.backend.regions.get(name)
        return Region(_backend=self.backend, _model=model)

    def delete(self, x: Region | int | str):
        if isinstance(x, Region):
            id = x.id
        elif isinstance(x, int):
            id = x
        elif isinstance(x, str):
            model = self.backend.regions.get(x)
            id = model.id
        else:
            raise TypeError("Invalid argument: Must be `Region`, `int` or `str`.")

        self.backend.regions.delete(id)

    def list(
        self,
        name: str | None = None,
        hierarchy: str | None = None,
    ) -> list[Region]:
        regions = self.backend.regions.list(name=name, hierarchy=hierarchy)
        return [Region(_backend=self.backend, _model=r) for r in regions]

    def tabulate(
        self,
        name: str | None = None,
        hierarchy: str | None = None,
    ) -> pd.DataFrame:
        return self.backend.regions.tabulate(name=name, hierarchy=hierarchy)

    def get_docs(self, name: str) -> str | None:
        region_id = self._get_region_id(name)
        if region_id is None:
            return None
        try:
            return self.backend.regions.docs.get(dimension_id=region_id).description
        except DocsModel.NotFound:
            return None

    def set_docs(self, name: str, description: str | None) -> str | None:
        if description is None:
            self.delete_docs(name=name)
            return None
        region_id = self._get_region_id(name)
        if region_id is None:
            return None
        return self.backend.regions.docs.set(
            dimension_id=region_id, description=description
        ).description

    def delete_docs(self, name: str) -> None:
        # TODO: this function is failing silently, which we should avoid
        region_id = self._get_region_id(name)
        if region_id is None:
            return None
        try:
            self.backend.regions.docs.delete(dimension_id=region_id)
            return None
        except DocsModel.NotFound:
            return None



# flake8: noqa

from .base import BaseModel, DataFrame
from .docs import Docs, DocsRepository
from .iamc import (  # Measurand,; MeasurandRepository,; AnnualDataPoint,; SubAnnualDataPoint,; CategoricalDataPoint,
    DataPoint,
    DataPointRepository,
    TimeSeries,
    TimeSeriesRepository,
    Variable,
    VariableRepository,
)
from .meta import RunMetaEntry, RunMetaEntryRepository
from .model import Model, ModelRepository
from .optimization import (
    IndexSet,
    IndexSetRepository,
    Scalar,
    ScalarRepository,
    Table,
    TableRepository,
)
from .region import Region, RegionParent, RegionRepository
from .run import Run, RunRepository
from .scenario import Scenario, ScenarioRepository
from .unit import Unit, UnitRepository



from typing import Generic, TypeVar

import pandas as pd
from pydantic import BaseModel as PydanticBaseModel
from pydantic import ConfigDict, Field

from ixmp4.conf import settings
from ixmp4.data import api


class BaseModel(PydanticBaseModel):
    model_config = ConfigDict(from_attributes=True, arbitrary_types_allowed=True)


EnumeratedT = TypeVar("EnumeratedT")


class Pagination(BaseModel):
    limit: int = Field(
        default=settings.default_page_size,
        ge=0,
        le=settings.max_page_size,
    )
    offset: int = Field(default=0, ge=0)


class EnumerationOutput(BaseModel, Generic[EnumeratedT]):
    pagination: Pagination
    total: int
    results: api.DataFrame | list[EnumeratedT]

    def __init__(
        __pydantic_self__,
        *args,
        results: pd.DataFrame | api.DataFrame | list[EnumeratedT],
        **kwargs,
    ):
        if isinstance(results, pd.DataFrame):
            kwargs["results"] = api.DataFrame.model_validate(results)
        else:
            kwargs["results"] = results
        super().__init__(**kwargs)



from collections import UserDict
from typing import ClassVar

import numpy as np
import pandas as pd

from ixmp4.data.abstract import Run as RunModel

from .base import BaseFacade, BaseModelFacade
from .iamc import RunIamcData
from .optimization import OptimizationData


class Run(BaseModelFacade):
    _model: RunModel
    _meta: "RunMetaFacade"
    NoDefaultVersion: ClassVar = RunModel.NoDefaultVersion
    NotFound: ClassVar = RunModel.NotFound
    NotUnique: ClassVar = RunModel.NotUnique

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)

        self.version = self._model.version

        self.iamc = RunIamcData(_backend=self.backend, run=self._model)
        self._meta = RunMetaFacade(_backend=self.backend, run=self._model)
        self.optimization = OptimizationData(_backend=self.backend, run=self._model)

    @property
    def model(self):
        """Associated model."""
        return self._model.model

    @property
    def scenario(self):
        """Associated scenario."""
        return self._model.scenario

    @property
    def id(self):
        """Unique id."""
        return self._model.id

    @property
    def meta(self):
        "Meta indicator data (`dict`-like)."
        return self._meta

    @meta.setter
    def meta(self, meta):
        self._meta._set(meta)

    def set_as_default(self):
        """Sets this run as the default version for its `model` + `scenario`
        combination."""
        self.backend.runs.set_as_default_version(self._model.id)

    def unset_as_default(self):
        """Unsets this run as the default version."""
        self.backend.runs.unset_as_default_version(self._model.id)


class RunRepository(BaseFacade):
    def create(
        self,
        model: str,
        scenario: str,
    ) -> Run:
        return Run(
            _backend=self.backend, _model=self.backend.runs.create(model, scenario)
        )

    def get(
        self,
        model: str,
        scenario: str,
        version: int | None = None,
    ) -> Run:
        if version is None:
            _model = self.backend.runs.get_default_version(model, scenario)
        else:
            _model = self.backend.runs.get(model, scenario, version)
        return Run(_backend=self.backend, _model=_model)

    def list(self, default_only: bool = True, **kwargs) -> list[Run]:
        return [
            Run(_backend=self.backend, _model=r)
            for r in self.backend.runs.list(default_only=default_only, **kwargs)
        ]

    def tabulate(self, default_only: bool = True, **kwargs) -> pd.DataFrame:
        runs = self.backend.runs.tabulate(default_only=default_only, **kwargs)
        runs["model"] = runs["model__id"].map(self.backend.models.map())
        runs["scenario"] = runs["scenario__id"].map(self.backend.scenarios.map())
        return runs[["id", "model", "scenario", "version", "is_default"]]


class RunMetaFacade(BaseFacade, UserDict):
    run: RunModel

    def __init__(self, run: RunModel, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.run = run
        self.df, self.data = self._get()

    def _get(self) -> tuple[pd.DataFrame, dict]:
        df = self.backend.meta.tabulate(run_id=self.run.id, run={"default_only": False})
        if df.empty:
            return df, {}
        return df, dict(zip(df["key"], df["value"]))

    def _set(self, meta: dict):
        df = pd.DataFrame({"key": self.data.keys()})
        df["run__id"] = self.run.id
        self.backend.meta.bulk_delete(df)
        df = pd.DataFrame(
            {"key": meta.keys(), "value": [numpy_to_pytype(v) for v in meta.values()]}
        )
        df.dropna(axis=0, inplace=True)
        df["run__id"] = self.run.id
        self.backend.meta.bulk_upsert(df)
        self.df, self.data = self._get()

    def __setitem__(self, key, value: int | float | str | bool):
        try:
            del self[key]
        except KeyError:
            pass

        value = numpy_to_pytype(value)
        if value is not None:
            self.backend.meta.create(self.run.id, key, value)
        self.df, self.data = self._get()

    def __delitem__(self, key):
        id = dict(zip(self.df["key"], self.df["id"]))[key]
        self.backend.meta.delete(id)
        self.df, self.data = self._get()


def numpy_to_pytype(value):
    """Cast numpy-types to basic Python types"""
    if value is np.nan:  # np.nan is cast to 'float', not None
        return None
    elif isinstance(value, np.generic):
        return value.item()
    else:
        return value



import asyncio
import logging
from json.decoder import JSONDecodeError
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    ClassVar,
    Coroutine,
    Generic,
    Type,
    TypeVar,
)

import httpx
import pandas as pd
from pydantic import BaseModel as PydanticBaseModel
from pydantic import ConfigDict, Field, GetCoreSchemaHandler
from pydantic_core import CoreSchema, core_schema

from ixmp4.conf import settings
from ixmp4.core.exceptions import (
    ApiEncumbered,
    ImproperlyConfigured,
    IxmpError,
    UnknownApiError,
    registry,
)

if TYPE_CHECKING:
    from ixmp4.data.backend.api import RestBackend

logger = logging.getLogger(__name__)


class BaseModel(PydanticBaseModel):
    NotFound: ClassVar[type[IxmpError]]
    NotUnique: ClassVar[type[IxmpError]]

    model_config = ConfigDict(from_attributes=True)


def df_to_dict(df: pd.DataFrame) -> dict:
    columns = []
    dtypes = []
    for c in df.columns:
        columns.append(c)
        dtypes.append(df[c].dtype.name)

    return {
        "index": df.index.to_list(),
        "columns": columns,
        "dtypes": dtypes,
        "data": df.values.tolist(),
    }


class DataFrame(PydanticBaseModel):
    index: list | None = Field(None)
    columns: list[str] | None
    dtypes: list[str] | None
    data: list | None

    model_config = ConfigDict(json_encoders={pd.Timestamp: lambda x: x.isoformat()})

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source_type: Any, handler: GetCoreSchemaHandler
    ) -> CoreSchema:
        return core_schema.no_info_before_validator_function(cls.validate, handler(cls))
        # yield cls.validate

    @classmethod
    def validate(cls, df: pd.DataFrame | dict):
        if isinstance(df, pd.DataFrame):
            return df_to_dict(df)
        else:
            return df

    def to_pandas(self) -> pd.DataFrame:
        df = pd.DataFrame(
            index=self.index or None,
            columns=self.columns,
            data=self.data,
        )
        if self.columns and self.dtypes:
            for c, dt in zip(self.columns, self.dtypes):
                # there seems to be a type incompatbility between StrDtypeArg and str
                df[c] = df[c].astype(dt)  # type: ignore
        return df


ModelType = TypeVar("ModelType", bound=BaseModel)


class BaseRepository(Generic[ModelType]):
    model_class: Type[ModelType]
    prefix: ClassVar[str]
    enumeration_method: str = "PATCH"

    backend: "RestBackend"

    def __init__(self, backend: "RestBackend", *args, **kwargs) -> None:
        self.backend = backend

    def sanitize_params(self, params: dict):
        return {k: params[k] for k in params if params[k] is not None}

    def get_remote_exception(self, res: httpx.Response, status_code: int):
        try:
            json = res.json()
        except (ValueError, JSONDecodeError):
            raise UnknownApiError(content=res.text, response_status=res.status_code)

        try:
            error_name = json["error_name"]
        except KeyError:
            raise UnknownApiError(json_data=json, response_status=status_code)

        try:
            exc = registry[error_name]
        except KeyError:
            raise ImproperlyConfigured(
                "Could not find remote exception in registry. "
                "Are you sure client and server ixmp versions are compatible?"
            )
        return exc.from_dict(json)

    def _request(self, *args, **kwargs) -> dict | list | None:
        res = asyncio.run(self._async_request(*args, **kwargs))
        return res

    async def _async_request(
        self,
        method: str,
        path: str,
        params: dict | None = None,
        json: dict | None = None,
        max_retries: int = settings.client_max_request_retries,
        **kwargs,
    ) -> dict | list | None:
        """Sends an asyncronous request and handles potential error responses.
        Uses the backend's semaphore to limit how many
        requests are sent concurrently per backend.
        Re-raises a remote `IxmpError` if thrown and transferred from the backend.
        Handles read timeouts and rate limiting responses
        via retries with exponential backoffs.
        Returns `None` if the response body is empty
        but has a status code less than 300.
        """

        async def retry(max_retries=max_retries) -> dict | list | None:
            if max_retries == 0:
                self.backend.semaphore
                logger.error(f"API Encumbered: '{self.backend.info.dsn}'")
                raise ApiEncumbered(
                    f"The service connected to the backend '{self.backend.info.name}' "
                    "is currently overencumbered with requests. "
                    "Try again at a later time or re-configure your client "
                    "to behave more leniently."
                )
            # increase backoff by `settings.client_backoff_factor`
            # for each retry
            backoff_s = settings.client_backoff_factor * (
                settings.client_max_request_retries - max_retries
            )
            logger.debug(f"Retrying request in {backoff_s} seconds.")
            await asyncio.sleep(backoff_s)
            return await self._async_request(
                method,
                path,
                params=params,
                json=json,
                max_retries=max_retries - 1,
                **kwargs,
            )

        if params is None:
            params = {}
        else:
            params = self.sanitize_params(params)

        try:
            async with self.backend.semaphore:
                res = await self.backend.async_client.request(
                    method, path, params=params, json=json, **kwargs
                )
        except httpx.ReadTimeout:
            logger.warn("Read timeout, retrying request...")
            return await retry()

        return await self._async_handle_response(res, retry)

    async def _async_handle_response(
        self,
        res: httpx.Response,
        retry: Callable[..., Coroutine[Any, Any, dict | list | None]],
    ) -> dict | list | None:
        if res.status_code in [
            429,  # Too Many Requests
            420,  # Enhance Your Calm
        ]:
            return await retry()
        elif res.status_code >= 400:
            if res.status_code == 413:
                raise ImproperlyConfigured(
                    "Received status code 413 (Payload Too Large). "
                    "Consider decreasing `IXMP4_CLIENT_DEFAULT_UPLOAD_CHUNK_SIZE` "
                    f"(current: {settings.client_default_upload_chunk_size})."
                )
            raise self.get_remote_exception(res, res.status_code)
        else:
            try:
                return res.json()
            except JSONDecodeError:
                if res.status_code < 300 and res.text == "":
                    return None
            except ValueError:
                pass
            raise UnknownApiError(res.text)

    def _get_by_id(self, id: int, *args, **kwargs) -> dict[str, Any]:
        # we can assume this type on create endpoints
        return self._request("GET", self.prefix + str(id) + "/", **kwargs)  # type: ignore

    def _request_enumeration(
        self,
        table: bool = False,
        params: dict | None = None,
        json: dict | None = None,
    ):
        """Convenience method for requests to the enumeration endpoint."""
        if params is None:
            params = {}

        return self._request(
            self.enumeration_method,
            self.prefix,
            params={**params, "table": table},
            json=json,
        )

    def _build_pagination_requests(
        self,
        total: int,
        start: int,
        limit: int,
        params: dict | None,
        json: dict | None,
    ) -> list[Coroutine]:
        requests: list[Coroutine] = []
        for req_offset in range(start, total, limit):
            if params is not None:
                req_params = params.copy()
            else:
                req_params = {}

            req_params.update({"limit": limit, "offset": req_offset})
            request = self._async_request(
                self.enumeration_method,
                self.prefix,
                params=req_params,
                json=json,
            )
            requests.append(request)
        return requests

    def _collect_pagination_results(self, requests: list[Coroutine]) -> list:
        async def gather():
            return await asyncio.gather(*requests)

        responses = asyncio.run(gather())
        return [r.pop("results") for r in responses]

    def _handle_pagination(
        self,
        data: dict,
        table: bool = False,
        params: dict | None = None,
        json: dict | None = None,
    ) -> list[list] | list[dict]:
        """Handles paginated response and sends subsequent requests if necessary.
        Returns aggregated pages as a list."""

        if params is None:
            params = {"table": table}
        else:
            params["table"] = table

        total = data.pop("total")
        pagination = data.pop("pagination")
        offset = pagination.pop("offset")
        limit = pagination.pop("limit")
        if total <= offset + limit:
            return [data.pop("results")]
        else:
            requests = self._build_pagination_requests(
                total, offset + limit, limit, params, json
            )
            results = self._collect_pagination_results(requests)
            return [data.pop("results")] + results

    def _list(
        self, params: dict | None = None, json: dict | None = None, **kwargs
    ) -> list[ModelType]:
        data = self._request_enumeration(params=params, table=False, json=json)
        if isinstance(data, dict):
            # we can assume this type on list endpoints
            pages: list[list] = self._handle_pagination(
                data, table=False, params=params, json=json
            )  # type: ignore
            results = [i for page in pages for i in page]
        else:
            results = data
        return [self.model_class(**i) for i in results]

    def _tabulate(
        self, params: dict | None = {}, json: dict | None = None, **kwargs
    ) -> pd.DataFrame:
        data = self._request_enumeration(table=True, params=params, json=json)
        pagination = data.get("pagination", None)
        if pagination is not None:
            # we can assume this type on table endpoints
            pages: list[dict] = self._handle_pagination(
                data,
                table=True,
                params=params,
                json=json,
            )  # type: ignore
            dfs = [DataFrame(**page).to_pandas() for page in pages]
            return pd.concat(dfs)
        else:
            return DataFrame(**data).to_pandas()

    def _create(self, *args, **kwargs) -> dict[str, Any]:
        # we can assume this type on create endpoints
        return self._request("POST", *args, **kwargs)  # type: ignore

    def _delete(self, id: int):
        self._request("DELETE", f"{self.prefix}{str(id)}/")


class Retriever(BaseRepository[ModelType]):
    def get(self, **kwargs) -> ModelType:
        if self.enumeration_method == "GET":
            list_ = self._list(params=kwargs)
        else:
            list_ = self._list(json=kwargs)

        try:
            [obj] = list_
        except ValueError as e:
            raise self.model_class.NotFound(
                f"Expected exactly one result, got {len(list_)} instead."
            ) from e
        return obj


class Creator(BaseRepository[ModelType]):
    def create(self, **kwargs) -> ModelType:
        res = self._create(
            self.prefix,
            json=kwargs,
        )
        return self.model_class(**res)


class Deleter(BaseRepository[ModelType]):
    def delete(self, id: int):
        self._delete(id)


class Lister(BaseRepository[ModelType]):
    def list(self, *args, **kwargs) -> list[ModelType]:
        return self._list(json=kwargs)


class Tabulator(BaseRepository[ModelType]):
    def tabulate(self, *args, **kwargs) -> pd.DataFrame:
        return self._tabulate(json=kwargs)


class Enumerator(Lister[ModelType], Tabulator[ModelType]):
    def enumerate(
        self, *args, table: bool = False, **kwargs
    ) -> list[ModelType] | pd.DataFrame:
        if table:
            return self.tabulate(*args, **kwargs)
        else:
            return self.list(*args, **kwargs)


class BulkOperator(BaseRepository[ModelType]):
    def yield_chunks(self, df: pd.DataFrame, chunk_size: int):
        for _, chunk in df.groupby(df.index // chunk_size):
            yield chunk


class BulkUpserter(BulkOperator[ModelType]):
    def bulk_upsert(
        self,
        df: pd.DataFrame,
        chunk_size: int = settings.client_default_upload_chunk_size,
        **kwargs,
    ):
        for chunk in self.yield_chunks(df, chunk_size):
            self.bulk_upsert_chunk(chunk, **kwargs)

    def bulk_upsert_chunk(self, df: pd.DataFrame, **kwargs) -> None:
        dict_ = df_to_dict(df)
        json_ = DataFrame(**dict_).model_dump_json()
        self._request(
            "POST",
            self.prefix + "bulk/",
            params=kwargs,
            content=json_,
        )


class BulkDeleter(BulkOperator[ModelType]):
    def bulk_delete(
        self,
        df: pd.DataFrame,
        chunk_size: int = settings.client_default_upload_chunk_size,
        **kwargs,
    ):
        for chunk in self.yield_chunks(df, chunk_size):
            self.bulk_delete_chunk(chunk, **kwargs)

    def bulk_delete_chunk(self, df: pd.DataFrame, **kwargs) -> None:
        dict_ = df_to_dict(df)
        json_ = DataFrame(**dict_).model_dump_json()
        self._request(
            "PATCH",
            self.prefix + "bulk/",
            params=kwargs,
            content=json_,
        )



from datetime import datetime
from typing import ClassVar

import pandas as pd

from ixmp4.data import abstract

from .. import base


class DataPoint(base.BaseModel):
    NotFound: ClassVar = abstract.DataPoint.NotFound
    NotUnique: ClassVar = abstract.DataPoint.NotUnique
    DeletionPrevented: ClassVar = abstract.DataPoint.DeletionPrevented

    id: int
    time_series__id: int
    value: float
    type: str

    step_category: str | None
    step_year: int | None
    step_datetime: datetime | None


class DataPointRepository(
    base.Enumerator[DataPoint],
    base.BulkUpserter[DataPoint],
    base.BulkDeleter[DataPoint],
    abstract.DataPointRepository,
):
    model_class = DataPoint
    prefix = "iamc/datapoints/"

    def enumerate(self, **kwargs) -> list[DataPoint] | pd.DataFrame:
        return super().enumerate(**kwargs)

    def list(
        self,
        join_parameters: bool | None = None,
        join_runs: bool | None = None,
        **kwargs,
    ) -> list[DataPoint]:
        return super()._list(
            json=kwargs,
            params={
                "join_parameters": join_parameters,
                "join_runs": join_runs,
            },
        )

    def tabulate(
        self,
        join_parameters: bool | None = None,
        join_runs: bool | None = None,
        **kwargs,
    ) -> pd.DataFrame:
        return super()._tabulate(
            json=kwargs,
            params={
                "join_parameters": join_parameters,
                "join_runs": join_runs,
            },
        )

    def bulk_upsert(self, df: pd.DataFrame, **kwargs) -> None:
        super().bulk_upsert(df)

    def bulk_delete(self, df: pd.DataFrame, **kwargs) -> None:
        super().bulk_delete(df)



from datetime import datetime
from typing import Any, ClassVar, Iterable

import pandas as pd

from ixmp4.data import abstract

from .. import base
from ..docs import Docs, DocsRepository
from .column import Column


class Table(base.BaseModel):
    NotFound: ClassVar = abstract.Table.NotFound
    NotUnique: ClassVar = abstract.Table.NotUnique
    DeletionPrevented: ClassVar = abstract.Table.DeletionPrevented

    id: int
    name: str
    data: dict[str, Any]
    columns: list["Column"]
    run__id: int

    created_at: datetime | None
    created_by: str | None


class TableDocsRepository(DocsRepository):
    model_class = Docs
    prefix = "docs/optimization/tables/"


class TableRepository(
    base.Creator[Table],
    base.Retriever[Table],
    base.Enumerator[Table],
    abstract.TableRepository,
):
    model_class = Table
    prefix = "optimization/tables/"

    def __init__(self, backend, *args, **kwargs) -> None:
        super().__init__(backend, *args, **kwargs)
        self.docs = TableDocsRepository(backend)

    def create(
        self,
        run_id: int,
        name: str,
        constrained_to_indexsets: list[str],
        column_names: list[str] | None = None,
    ) -> Table:
        return super().create(
            name=name,
            run_id=run_id,
            constrained_to_indexsets=constrained_to_indexsets,
            column_names=column_names,
        )

    def add_data(self, table_id: int, data: dict[str, Any] | pd.DataFrame) -> None:
        if isinstance(data, pd.DataFrame):
            # data will always contains str, not only Hashable
            data: dict[str, Any] = data.to_dict(orient="list")  # type: ignore
        kwargs = {"data": data}
        self._request(
            method="PATCH", path=self.prefix + str(table_id) + "/data/", json=kwargs
        )

    def get(self, run_id: int, name: str) -> Table:
        return super().get(run_id=run_id, name=name)

    def get_by_id(self, id: int) -> Table:
        res = self._get_by_id(id)
        return Table(**res)

    def list(self, *args, **kwargs) -> Iterable[Table]:
        return super().list(*args, **kwargs)

    def tabulate(self, *args, **kwargs) -> pd.DataFrame:
        return super().tabulate(*args, **kwargs)

    def enumerate(self, *args, **kwargs) -> Iterable[Table] | pd.DataFrame:
        return super().enumerate(*args, **kwargs)



from ixmp4.data import abstract
from ixmp4.data.auth.decorators import guard

from .. import base
from ..docs import BaseDocsRepository, docs_model
from .model import Unit

UnitDocs = docs_model(Unit)


class UnitDocsRepository(
    BaseDocsRepository,
    base.BaseRepository,
):
    model_class = UnitDocs
    dimension_model_class = Unit

    @guard("view")
    def list(self, *, dimension_id: int | None = None) -> list[abstract.Docs]:
        return super().list(dimension_id=dimension_id)



from datetime import datetime
from typing import Any

from sqlalchemy.orm import Mapped as Mapped

Boolean = Mapped[bool]
DateTime = Mapped[datetime]
Float = Mapped[float]
Integer = Mapped[int]
JsonList = Mapped[list[float | int | str]]
JsonDict = Mapped[dict[str, Any]]
String = Mapped[str]



import json
from typing import Optional

import typer
import uvicorn  # type: ignore[import]
from fastapi.openapi.utils import get_openapi

from ixmp4.conf import settings
from ixmp4.server import v1

from . import utils

app = typer.Typer()


@app.command()
def start(
    host: str = typer.Option(default="127.0.0.1", help="The hostname to bind to."),
    port: int = typer.Option(default=9000, help="Requested server port."),
) -> None:
    """Starts the ixmp4 web api."""
    reload = settings.mode != "production"
    uvicorn.run(
        "ixmp4.server:app",
        host=host,
        port=port,
        reload=reload,
        log_config="ixmp4/conf/logging/server.conf",
    )


@app.command()
def dump_schema(output_file: Optional[typer.FileTextWrite] = typer.Option(None, "-o")):
    schema = get_openapi(
        title=v1.title,
        version=v1.version,
        openapi_version=v1.openapi_version,
        description=v1.description,
        routes=v1.routes,
    )
    if output_file is None:
        utils.echo(json.dumps(schema))
    else:
        json.dump(
            schema,
            output_file,
        )



# flake8: noqa
from .iamc.variable import Variable as Variable
from .model import Model as Model
from .optimization.indexset import IndexSet as IndexSet
from .optimization.scalar import Scalar as Scalar
from .optimization.table import Table as Table
from .platform import Platform as Platform
from .region import Region as Region
from .run import Run as Run
from .scenario import Scenario as Scenario
from .unit import Unit as Unit



import logging
import time

from starlette.middleware.base import BaseHTTPMiddleware

logger = logging.getLogger(__name__)


class RequestTimeLoggerMiddleware(BaseHTTPMiddleware):
    async def dispatch(self, request, call_next):
        start_time = time.time()
        response = await call_next(request)
        process_time = time.time() - start_time
        logger.debug(f"Request process time: {process_time} seconds.")
        return response


class RequestSizeLoggerMiddleware(BaseHTTPMiddleware):
    async def dispatch(self, request, call_next):
        body = await request.body()
        logger.debug(f"Request body size: {len(body)} bytes.")
        return await call_next(request)



import pandas as pd


def map_existing(
    df: pd.DataFrame,
    existing_df: pd.DataFrame,
    join_on: tuple[str, str],
    map: tuple[str, str],
):
    _, join_to = join_on
    _, map_to = map
    existing_df = existing_df.rename(columns=dict([join_on, map]))[[join_to, map_to]]
    df = df.merge(existing_df, how="left", on=[join_to])
    missing = df.where(pd.isnull(df[map_to]))[join_to].dropna().unique()
    return df, missing



from typing import ClassVar, Protocol

import pandas as pd

from ixmp4.core.exceptions import IxmpError, NoDefaultRunVersion
from ixmp4.data import types

from . import base


class Run(base.BaseModel, Protocol):
    """Model run data model."""

    NoDefaultVersion: ClassVar[type[IxmpError]] = NoDefaultRunVersion

    model__id: types.Integer
    "Foreign unique integer id of the model."
    model: types.Mapped
    "Associated model."

    scenario__id: types.Integer
    "Foreign unique integer id of the scenario."
    scenario: types.Mapped
    "Associated scenario."

    version: types.Integer
    "Run version."
    is_default: types.Boolean
    "`True` if this is the default run version."

    def __str__(self) -> str:
        return f"<Run {self.id} model={self.model.name} \
            scenario={self.scenario.name} version={self.version} \
            is_default={self.is_default}>"


class RunRepository(
    base.Creator,
    base.Retriever,
    base.Enumerator,
    Protocol,
):
    def create(
        self,
        model_name: str,
        scenario_name: str,
    ) -> Run:
        """Creates a run with an incremented version number or version=1 if no versions
        exist. Will automatically create the models and scenarios if they don't exist
        yet.

        Parameters
        ----------
        model_name : str
            The name of a model.
        scenario_name : str
            The name of a scenario.

        Returns
        -------
        :class:`ixmp4.data.abstract.Run`:
            The created run.
        """
        ...

    def get(
        self,
        model_name: str,
        scenario_name: str,
        version: int,
    ) -> Run:
        """Retrieves a run.

        Parameters
        ----------
        model_name : str
            The name of a model.
        scenario_name : str
            The name of a scenario.
        version : int
            The version number of this run.

        Raises
        ------
        :class:`ixmp4.data.abstract.Run.NotFound`:
            If the run with `model_name`, `scenario_name` and `version` does not exist.

        Returns
        -------
        :class:`ixmp4.data.abstract.Run`:
            The retrieved run.
        """
        ...

    def get_or_create(
        self,
        model_name: str,
        scenario_name: str,
    ) -> Run:
        """Tries to retrieve a run's default version
        and creates it if it was not found.

        Parameters
        ----------
        model_name : str
            The name of a model.
        scenario_name : str
            The name of a scenario.

        Returns
        -------
        :class:`ixmp4.data.abstract.Run`:
            The retrieved or created run.
        """
        try:
            return self.get_default_version(model_name, scenario_name)
        except Run.NoDefaultVersion:
            return self.create(model_name, scenario_name)

    def get_default_version(
        self,
        model_name: str,
        scenario_name: str,
    ) -> Run:
        """Retrieves a run's default version.

        Parameters
        ----------
        model_name : str
            The name of a model.
        scenario_name : str
            The name of a scenario.

        Raises
        ------
        :class:`ixmp4.core.exceptions.NoDefaultRunVersion`:
            If no runs with `model_name`, `scenario_name` and `is_default=True` exist.

        Returns
        -------
        :class:`ixmp4.data.abstract.Run`:
            The retrieved run.
        """
        ...

    def list(
        self,
        *,
        version: int | None = None,
        **kwargs,
    ) -> list[Run]:
        r"""Lists runs by specified criteria.

        Parameters
        ----------
        version : int
            The run's version.
        default_only : bool
            True by default. This function will return default runs only if true.
        \*\*kwargs: any
            More filter parameters as specified in
            `ixmp4.data.db.run.filters.RunFilter`.

        Returns
        -------
        Iterable[:class:`ixmp4.data.abstract.Run`]:
            List of runs.
        """
        ...

    def tabulate(
        self,
        *,
        version: int | None = None,
        **kwargs,
    ) -> pd.DataFrame:
        r"""Tabulate runs by specified criteria.

        Parameters
        ----------
        version : int
            The run's version.
        default_only : bool
            True by default. This function will return default runs only if true.
        \*\*kwargs: any
            More filter parameters as specified in
            `ixmp4.data.db.run.filters.RunFilter`.

        Returns
        -------
        :class:`pandas.DataFrame`:
            A data frame with the columns:
                - id
                - model__id
                - scenario__id
        """
        ...

    def set_as_default_version(self, id: int) -> None:
        """Sets a run as the default version for a (model, scenario) combination.

        Parameters
        ----------
        id : int
            Unique integer id.

        Raises
        ------
        :class:`ixmp4.data.abstract.Run.NotFound`:
            If no run with the `id` exists.

        """
        ...

    def unset_as_default_version(self, id: int) -> None:
        """Unsets a run as the default version leaving no
        default version for a (model, scenario) combination.

        Parameters
        ----------
        id : int
            Unique integer id.

        Raises
        ------
        :class:`ixmp4.data.abstract.Run.NotFound`:
            If no run with the `id` exists.
        :class:`ixmp4.core.exceptions.IxmpError`:
            If the run is not set as a default version.

        """
        ...



from typing import ClassVar

from ixmp4.db import filters

from .. import Region, TimeSeries


class RegionFilter(filters.BaseFilter, metaclass=filters.FilterMeta):
    id: filters.Id
    name: filters.String
    hierarchy: filters.String

    sqla_model: ClassVar[type] = Region

    def join(self, exc, **kwargs):
        exc = exc.join(Region, TimeSeries.region)
        return exc



import pandas as pd

from ixmp4 import db
from ixmp4.data import abstract
from ixmp4.data.auth.decorators import guard

from .. import base
from .docs import ModelDocsRepository
from .model import Model


class ModelRepository(
    base.Creator[Model],
    base.Retriever[Model],
    base.Enumerator[Model],
    abstract.ModelRepository,
):
    model_class = Model
    docs: ModelDocsRepository

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.docs = ModelDocsRepository(*args, **kwargs)

        from .filter import ModelFilter

        self.filter_class = ModelFilter

    def add(self, name: str) -> Model:
        model = Model(name=name, **self.get_creation_info())
        self.session.add(model)
        return model

    @guard("view")
    def get(self, name: str) -> Model:
        exc = self.select(name=name)
        try:
            return self.session.execute(exc).scalar_one()
        except db.NoResultFound:
            raise Model.NotFound

    @guard("edit")
    def create(self, *args, **kwargs) -> Model:
        return super().create(*args, **kwargs)

    @guard("view")
    def list(self, *args, **kwargs) -> list[Model]:
        return super().list(*args, **kwargs)

    @guard("view")
    def tabulate(self, *args, **kwargs) -> pd.DataFrame:
        return super().tabulate(*args, **kwargs)



import pandas as pd
from sqlalchemy.exc import NoResultFound

from ixmp4 import db
from ixmp4.core.exceptions import Forbidden, IxmpError, NoDefaultRunVersion
from ixmp4.data import abstract
from ixmp4.data.auth.decorators import guard
from ixmp4.db import utils

from .. import base
from ..model import Model, ModelRepository
from ..scenario import Scenario, ScenarioRepository
from .model import Run


class RunRepository(
    base.Creator[Run],
    base.Retriever[Run],
    base.Enumerator[Run],
    abstract.RunRepository,
):
    model_class = Run

    models: ModelRepository
    scenarios: ScenarioRepository

    def __init__(self, *args, **kwargs) -> None:
        self.models = ModelRepository(*args, **kwargs)
        self.scenarios = ScenarioRepository(*args, **kwargs)

        from .filter import RunFilter

        self.filter_class = RunFilter
        super().__init__(*args, **kwargs)

    def join_auth(self, exc: db.sql.Select):
        if not utils.is_joined(exc, Model):
            exc = exc.join(Model, Run.model)
        return exc

    def add(self, model_name: str, scenario_name: str) -> Run:
        # Get or create model
        try:
            exc: db.sql.Select = self.models.select(name=model_name)
            model = self.session.execute(exc).scalar_one()
        except NoResultFound:
            model = Model(name=model_name, **self.get_creation_info())
            self.session.add(model)

        # Get or create scenario
        try:
            exc = self.scenarios.select(name=scenario_name)
            scenario: Scenario = self.session.execute(exc).scalar_one()
        except NoResultFound:
            scenario = Scenario(name=scenario_name, **self.get_creation_info())
            self.session.add(scenario)

        (version,) = (
            # Aggregate MAX over run.version where the run's
            # model and scenario match.
            self.session.query(db.func.max(Run.version))
            .filter(Run.model.has(name=model_name))
            .filter(Run.scenario.has(name=scenario_name))
            .one()
        )
        if version is None:
            version = 0
        version += 1

        run = Run(model=model, scenario=scenario, version=version)
        self.session.add(run)
        return run

    @guard("edit")
    def create(self, model_name: str, *args, **kwargs) -> Run:
        if self.backend.auth_context is not None:
            if not self.backend.auth_context.check_access("edit", model_name):
                raise Forbidden(f"Access to model '{model_name}' denied.")
        return super().create(model_name, *args, **kwargs)

    @guard("view")
    def get(
        self,
        model_name: str,
        scenario_name: str,
        version: int,
    ) -> Run:
        exc = self.select(
            model={"name": model_name},
            scenario={"name": scenario_name},
            version=version,
            default_only=False,
        )

        try:
            return self.session.execute(exc).scalar_one()
        except NoResultFound:
            raise Run.NotFound(
                model=model_name,
                scenario=scenario_name,
                version=version,
            )

    @guard("view")
    def get_by_id(self, id: int) -> Run:
        obj = self.session.get(self.model_class, id)

        if obj is None:
            raise Run.NotFound(id=id)

        return obj

    @guard("view")
    def get_default_version(
        self,
        model_name: str,
        scenario_name: str,
    ) -> Run:
        exc = self.select(
            model={"name": model_name},
            scenario={"name": scenario_name},
        )

        try:
            return self.session.execute(exc).scalar_one()
        except NoResultFound:
            raise NoDefaultRunVersion

    @guard("view")
    def tabulate(self, *args, **kwargs) -> pd.DataFrame:
        return super().tabulate(*args, **kwargs)

    @guard("view")
    def list(self, *args, **kwargs) -> list[Run]:
        return super().list(*args, **kwargs)

    @guard("edit")
    def set_as_default_version(self, id: int) -> None:
        try:
            run = self.session.get(Run, id)
        except NoResultFound:
            raise Run.NotFound(id=id)

        if run is None:
            raise Run.NotFound

        exc = (
            db.update(Run)
            .where(
                Run.model__id == run.model__id,
                Run.scenario__id == run.scenario__id,
            )
            .values(is_default=False)
        )

        self.session.execute(exc)
        self.session.commit()

        run.is_default = True
        self.session.commit()

    @guard("edit")
    def unset_as_default_version(self, id: int) -> None:
        try:
            run = self.session.get(Run, id)
        except NoResultFound:
            raise Run.NotFound(id=id)

        if run is None:
            raise Run.NotFound

        if not run.is_default:
            raise IxmpError(f"Run with id={id} is not set as the default version.")

        run.is_default = False
        self.session.commit()



from datetime import datetime
from typing import Any, ClassVar, Iterable

import pandas as pd

from ixmp4.core.base import BaseFacade, BaseModelFacade
from ixmp4.data.abstract import Docs as DocsModel
from ixmp4.data.abstract import Run
from ixmp4.data.abstract import Table as TableModel
from ixmp4.data.abstract.optimization import Column


class Table(BaseModelFacade):
    _model: TableModel
    NotFound: ClassVar = TableModel.NotFound
    NotUnique: ClassVar = TableModel.NotUnique

    @property
    def id(self) -> int:
        return self._model.id

    @property
    def name(self) -> str:
        return self._model.name

    @property
    def run_id(self) -> int:
        return self._model.run__id

    @property
    def data(self) -> dict[str, Any]:
        return self._model.data

    def add(self, data: dict[str, Any] | pd.DataFrame) -> None:
        """Adds data to an existing Table."""
        self.backend.optimization.tables.add_data(table_id=self._model.id, data=data)
        self._model.data = self.backend.optimization.tables.get(
            run_id=self._model.run__id, name=self._model.name
        ).data

    @property
    def constrained_to_indexsets(self) -> list[str]:
        return [column.indexset.name for column in self._model.columns]

    @property
    def columns(self) -> list[Column]:
        return self._model.columns

    @property
    def created_at(self) -> datetime | None:
        return self._model.created_at

    @property
    def created_by(self) -> str | None:
        return self._model.created_by

    @property
    def docs(self):
        try:
            return self.backend.optimization.tables.docs.get(self.id).description
        except DocsModel.NotFound:
            return None

    @docs.setter
    def docs(self, description):
        if description is None:
            self.backend.optimization.tables.docs.delete(self.id)
        else:
            self.backend.optimization.tables.docs.set(self.id, description)

    @docs.deleter
    def docs(self):
        try:
            self.backend.optimization.tables.docs.delete(self.id)
        # TODO: silently failing
        except DocsModel.NotFound:
            return None

    def __str__(self) -> str:
        return f"<Table {self.id} name={self.name}>"


class TableRepository(BaseFacade):
    _run: Run

    def __init__(self, _run: Run, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._run = _run

    def create(
        self,
        name: str,
        constrained_to_indexsets: list[str],
        column_names: list[str] | None = None,
    ) -> Table:
        model = self.backend.optimization.tables.create(
            name=name,
            run_id=self._run.id,
            constrained_to_indexsets=constrained_to_indexsets,
            column_names=column_names,
        )
        return Table(_backend=self.backend, _model=model)

    def get(self, name: str) -> Table:
        model = self.backend.optimization.tables.get(run_id=self._run.id, name=name)
        return Table(_backend=self.backend, _model=model)

    def list(self, name: str | None = None) -> Iterable[Table]:
        tables = self.backend.optimization.tables.list(run_id=self._run.id, name=name)
        return [
            Table(
                _backend=self.backend,
                _model=i,
            )
            for i in tables
        ]

    def tabulate(self, name: str | None = None) -> pd.DataFrame:
        return self.backend.optimization.tables.tabulate(name=name)



from typing import Union

import pandas as pd
import pandera as pa
from pandera.typing import DataFrame, Series
from sqlalchemy.exc import NoResultFound

from ixmp4 import db
from ixmp4.core.decorators import check_types
from ixmp4.data import abstract
from ixmp4.data.auth.decorators import guard
from ixmp4.data.db.model import Model
from ixmp4.data.db.run import Run
from ixmp4.data.db.scenario import Scenario

from .. import base
from .model import RunMetaEntry


class RemoveRunMetaEntryFrameSchema(pa.DataFrameModel):
    key: Series[pa.String] = pa.Field(coerce=True)
    run__id: Series[pa.Int] = pa.Field(coerce=True)

    class Config:
        strict = True
        coerce = True


class AddRunMetaEntryFrameSchema(RemoveRunMetaEntryFrameSchema):
    value: Series[pa.Object] = pa.Field(coerce=True)


class UpdateRunMetaEntryFrameSchema(AddRunMetaEntryFrameSchema):
    id: Series[pa.Int] = pa.Field(coerce=True)


class RunMetaEntryRepository(
    base.Creator[RunMetaEntry],
    base.Enumerator[RunMetaEntry],
    base.BulkUpserter[RunMetaEntry],
    base.BulkDeleter[RunMetaEntry],
    abstract.RunMetaEntryRepository,
):
    model_class = RunMetaEntry

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        from .filter import RunMetaEntryFilter

        self.filter_class = RunMetaEntryFilter

    def add(
        self, run__id: int, key: str, value: Union[str, int, bool, float]
    ) -> RunMetaEntry:
        if self.backend.auth_context is not None:
            self.backend.runs.check_access(
                {run__id},
                access_type="edit",
                is_default=None,
                default_only=False,
            )

        entry = RunMetaEntry(run__id=run__id, key=key, value=value)
        self.session.add(entry)
        return entry

    def check_df_access(self, df: pd.DataFrame):
        if self.backend.auth_context is not None:
            ts_ids = set(df["run__id"].unique().tolist())
            self.backend.runs.check_access(
                ts_ids,
                access_type="edit",
                is_default=None,
                default_only=False,
            )

    @guard("edit")
    def create(self, *args, **kwargs) -> RunMetaEntry:
        return super().create(*args, **kwargs)

    @guard("view")
    def get(self, run__id: int, key: str) -> RunMetaEntry:
        if self.backend.auth_context is not None:
            self.backend.runs.check_access(
                {run__id},
                access_type="view",
                is_default=None,
                default_only=False,
            )

        exc = self.select(
            run_id=run__id,
            key=key,
        )

        try:
            return self.session.execute(exc).scalar_one()
        except NoResultFound:
            raise RunMetaEntry.NotFound(
                run__id=run__id,
                key=key,
            )

    @guard("edit")
    def delete(self, id: int) -> None:
        if self.backend.auth_context is not None:
            try:
                pre_exc = db.select(RunMetaEntry).where(RunMetaEntry.id == id)
                meta = self.session.execute(pre_exc).scalar_one()
            except NoResultFound:
                raise RunMetaEntry.NotFound(
                    id=id,
                )
            self.backend.runs.check_access(
                {meta.run__id},
                access_type="edit",
                is_default=None,
                default_only=False,
            )

        exc = db.delete(RunMetaEntry).where(RunMetaEntry.id == id)

        try:
            self.session.execute(exc)
            self.session.commit()
        except NoResultFound:
            raise RunMetaEntry.NotFound(
                id=id,
            )

    def join_auth(self, exc: db.sql.Select) -> db.sql.Select:
        if not db.utils.is_joined(exc, Run):
            exc = exc.join(Run, RunMetaEntry.run)
        if not db.utils.is_joined(exc, Model):
            exc = exc.join(Model, Run.model)

        return super().join_auth(exc)

    def select_with_run_index(self) -> db.sql.Select:
        _exc = db.select(
            Model.name.label("model_name"),
            Scenario.name.label("scenario_name"),
            Run.version.label("version"),
            self.bundle,
        ).select_from(self.model_class)

        return (
            _exc.join(Run, Run.id == RunMetaEntry.run__id)
            .join(Model, onclause=Model.id == Run.model__id)
            .join(Scenario, onclause=Scenario.id == Run.scenario__id)
        )

    @guard("view")
    def list(self, *args, **kwargs) -> list[RunMetaEntry]:
        return super().list(*args, **kwargs)

    @guard("view")
    def tabulate(
        self,
        *args,
        join_run_index: bool = False,
        _raw: bool | None = False,
        **kwargs,
    ) -> pd.DataFrame:
        if _raw:
            return super().tabulate(*args, **kwargs)

        if join_run_index:
            _exc = self.select_with_run_index()
            df = super().tabulate(*args, _exc=_exc, **kwargs)
            df.drop(columns="run__id", inplace=True)
            df.rename(
                columns={"model_name": "model", "scenario_name": "scenario"},
                inplace=True,
            )
            index_columns = ["model", "scenario", "version"]
        else:
            df = super().tabulate(*args, **kwargs)
            index_columns = ["run__id"]

        if df.empty:
            return pd.DataFrame(
                [], columns=index_columns + ["id", "type", "key", "value"]
            )

        def map_value_column(df: pd.DataFrame):
            type_str = df.name
            type_ = RunMetaEntry.Type(type_str)
            col = RunMetaEntry._column_map[type_]
            df["value"] = df[col]
            df["type"] = type_str
            return df.drop(columns=RunMetaEntry._column_map.values())

        return df.groupby("type", group_keys=False).apply(
            map_value_column, include_groups=False
        )

    @check_types
    @guard("edit")
    def bulk_upsert(self, df: DataFrame[AddRunMetaEntryFrameSchema]) -> None:
        self.check_df_access(df)
        df["type"] = df["value"].map(type).map(RunMetaEntry.Type.from_pytype)

        type_: RunMetaEntry.Type
        for type_, type_df in df.groupby("type"):
            col = RunMetaEntry._column_map[type_]
            null_cols = set(RunMetaEntry._column_map.values()) - set([col])
            type_df["type"] = type_df["type"].map(lambda x: x.value)
            type_df = type_df.rename(columns={"value": col})

            # ensure all other columns are overwritten
            for nc in null_cols:
                type_df[nc] = None

            super().bulk_upsert(type_df)

    @check_types
    @guard("edit")
    def bulk_delete(self, df: DataFrame[RemoveRunMetaEntryFrameSchema]) -> None:
        self.check_df_access(df)
        return super().bulk_delete(df)



from typing import Protocol

import pandas as pd

from ixmp4.data import types

from . import base
from .docs import DocsRepository


class Model(base.BaseModel, Protocol):
    """Data model of an assement modeling "model".
    Unfortunately two naming conventions clash here.
    """

    name: types.String
    "Unique name of the model."

    created_at: types.DateTime
    "Creation date/time. TODO"
    created_by: types.String
    "Creator. TODO"

    def __str__(self) -> str:
        return f"<Model {self.id} name={self.name}>"


class ModelRepository(
    base.Creator,
    base.Retriever,
    base.Enumerator,
    Protocol,
):
    docs: DocsRepository

    def create(self, name: str) -> Model:
        """Creates a model.

        Parameters
        ----------
        name : str
            The name of the model.

        Raises
        ------
        :class:`ixmp4.core.exceptions.ModelNotUnique`:
            If the model with `name` is not unique.

        Returns
        -------
        :class:`ixmp4.data.abstract.Model`:
            The created model.
        """
        ...

    def get(self, name: str) -> Model:
        """Retrieves a model.

        Parameters
        ----------
        name : str
            The unique name of the model.

        Raises
        ------
        :class:`ixmp4.data.abstract.Model.NotFound`:
            If the model with `name` does not exist.

        Returns
        -------
        :class:`ixmp4.data.abstract.Model`:
            The retrieved model.
        """
        ...

    def list(
        self,
        *,
        name: str | None = None,
    ) -> list[Model]:
        """Lists models by specified criteria.

        Parameters
        ----------
        name : str
            The name of a model. If supplied only one result will be returned.

        Returns
        -------
        Iterable[:class:`ixmp4.data.abstract.Model`]:
            List of Model.
        """
        ...

    def tabulate(
        self,
        *,
        name: str | None = None,
    ) -> pd.DataFrame:
        """Tabulate models by specified criteria.

        Parameters
        ----------
        name : str
            The name of a model. If supplied only one result will be returned.

        Returns
        -------
        :class:`pandas.DataFrame`:
            A data frame with the columns:
                - id
                - name

        """
        ...

    def map(self, *args, **kwargs) -> dict:
        """Return a mapping of model-id to model-name.

        Returns
        -------
        :class:`dict`:
            A dictionary `id` -> `name`
        """
        return dict([(m.id, m.name) for m in self.list(*args, **kwargs)])



from typing import ClassVar

from ixmp4.db import filters, utils

from .. import Run
from ..iamc import TimeSeries
from .model import ModelFilter
from .scenario import ScenarioFilter


class RunFilter(filters.BaseFilter, metaclass=filters.FilterMeta):
    id: filters.Id
    version: filters.Integer
    default_only: filters.Boolean = filters.Field(True)
    is_default: filters.Boolean
    model: ModelFilter
    scenario: ScenarioFilter

    sqla_model: ClassVar[type] = Run

    def filter_default_only(self, exc, c, v, **kwargs):
        if v:
            return exc.where(Run.is_default)
        else:
            return exc

    def join(self, exc, **kwargs):
        if not utils.is_joined(exc, Run):
            exc = exc.join(Run, TimeSeries.run)
        return exc
